# -*- coding: utf-8 -*-

# built ins
import asyncio

# installed
from loguru import logger as log

# user defined formulas
from streaming_helper.db_management import sqlite_management as db_mgt
from streaming_helper.restful_api.deribit import end_point_params_template as end_point
from streaming_helper.utilities import (
    error_handling,
    pickling,
    string_modification as str_mod,
    system_tools,
    time_modification as time_mod,
)


async def initial_procedures(
    private_data: object,
    client_redis: object,
    config_app: list,
) -> None:

    try:

        # get tradable strategies
        tradable_config_app = config_app["tradable"]

        # get TRADABLE currencies
        currencies: list = [o["spot"] for o in tradable_config_app][0]

        # get ALL traded currencies in deribit
        get_currencies_all = await end_point.get_currencies()

        all_exc_currencies = [o["currency"] for o in get_currencies_all["result"]]

        server_time = time_mod.get_now_unix_time()

        ONE_SECOND = 1000

        one_minute = ONE_SECOND * 60

        five_days_ago = server_time - (one_minute * 60 * 24 * 5)

        my_path_cur = system_tools.provide_path_for_file("currencies")

        pickling.replace_data(
            my_path_cur,
            all_exc_currencies,
        )

        for currency in all_exc_currencies:

            # get ALL traded currencies in deribit
            instruments = await end_point.get_instruments(currency)

            my_path_instruments = system_tools.provide_path_for_file(
                "instruments", currency
            )

            pickling.replace_data(
                my_path_instruments,
                instruments,
            )


    except Exception as error:

        await error_handling.parse_error_message_with_redis(
            client_redis,
            error,
        )


def portfolio_combining(
    portfolio_all: list,
    portfolio_channel: str,
    result_template: dict,
) -> dict:

    portfolio = (
        [o["portfolio"] for o in portfolio_all if o["type"] == "subaccount"][0]
    ).values()

    #! need to update currency to upper

    result_template["params"].update({"data": portfolio})
    result_template["params"].update({"channel": portfolio_channel})

    return result_template


def my_trades_active_combining(
    my_trades_active_all: list,
    my_trades_channel: str,
    result_template: dict,
) -> dict:

    result_template["params"].update({"data": my_trades_active_all})
    result_template["params"].update({"channel": my_trades_channel})

    return result_template


def sub_account_combining(
    sub_accounts: list,
    sub_account_cached_channel: str,
    result_template: dict,
) -> dict:

    orders_cached = []
    positions_cached = []

    try:

        for sub_account in sub_accounts:

            sub_account = sub_account[0]

            sub_account_orders = sub_account["open_orders"]

            if sub_account_orders:

                for order in sub_account_orders:

                    orders_cached.append(order)

            sub_account_positions = sub_account["positions"]

            if sub_account_positions:

                for position in sub_account_positions:

                    positions_cached.append(position)

        sub_account = dict(
            orders_cached=orders_cached,
            positions_cached=positions_cached,
        )

        result_template["params"].update({"data": sub_account})
        result_template["params"].update({"channel": sub_account_cached_channel})

        return result_template

    except:

        sub_account = dict(
            orders_cached=[],
            positions_cached=[],
        )

        result_template["params"].update({"data": sub_account})
        result_template["params"].update({"channel": sub_account_cached_channel})

        return result_template


def is_order_allowed_combining(
    all_instruments_name: list,
    order_allowed_channel: str,
    result_template: dict,
) -> dict:

    combined_order_allowed = []
    for instrument_name in all_instruments_name:

        currency: str = str_mod.extract_currency_from_text(instrument_name)

        if "-FS-" in instrument_name:
            size_is_reconciled = 1

        else:
            size_is_reconciled = 0

        order_allowed = dict(
            instrument_name=instrument_name,
            size_is_reconciled=size_is_reconciled,
            currency=currency,
        )

        combined_order_allowed.append(order_allowed)

    result_template["params"].update({"data": combined_order_allowed})
    result_template["params"].update({"channel": order_allowed_channel})

    return result_template



# -*- coding: utf-8 -*-


# built ins
import asyncio

# installed
import orjson
from loguru import logger as log

# user defined formula
from streaming_helper.db_management import redis_client, sqlite_management as db_mgt
from streaming_helper.utilities import (
    error_handling,
    string_modification as str_mod,
)


"""

https://www.dataleadsfuture.com/exploring-numexpr-a-powerful-engine-behind-pandas/

    """


def ohlc_to_candlestick(conversion_array):

    candlestick_data = [0, 0, 0, 0, 0, 0]

    # log.info(f"conversion_array {conversion_array}")

    open = conversion_array[0]
    high = conversion_array[1]
    low = conversion_array[2]
    close = conversion_array[3]

    body_size = abs(close - open)
    height = abs(high - low)

    # log.debug(f"open {open} high {high} low {low} close {close}")
    # log.warning(f"body_size {body_size} height {height} close > open {close > open}")

    if close > open:
        candle_type = 1
        wicks_up = abs(high - close)
        wicks_down = abs(low - open)

    else:
        candle_type = -1
        wicks_up = abs(high - open)
        wicks_down = abs(low - close)

    candlestick_data[0] = candle_type

    candlestick_data[1] = round(round(wicks_up, 5), 2)

    candlestick_data[2] = round(round(wicks_down, 5), 2)

    candlestick_data[3] = round(round(body_size, 5), 2)

    candlestick_data[4] = round(round(height, 5), 2)

    candlestick_data[5] = (
        0 if body_size == 0 else (round(round(body_size / height, 5), 2) > 70 / 100) * 1
    )
    # log.info(f"candlestick_data {candlestick_data}")

    return candlestick_data


def my_generator_candle(
    np: object,
    data: list,
    lookback: int,
) -> list:
    """_summary_
        https://github.com/MikePapinski/DeepLearning/blob/master/PredictCandlestick/CandleSTick%20patterns%20prediction/JupyterResearch_0.1.ipynb
        https://mikepapinski.github.io/deep%20learning/machine%20learning/python/forex/2018/12/15/Predict-Candlestick-patterns-with-Keras-and-Forex-data.md.html

    Args:
        data (_type_): _description_
        lookback (_type_): _description_

    Returns:
        _type_: _description_
    """
    first_row = 0

    parameters = len(
        [
            "candle_type",
            "wicks_up",
            "wicks_down",
            "body_size",
            "length",
            "is_long_body",
        ]
    )

    # generating placeholder for random value
    arr = np.empty((1, lookback, parameters), int)

    # log.warning (f"data  {data}")
    # log.info (f"arr  {arr} ")
    # log.info (f"parameters  {parameters} lookback  {lookback}")

    for a in range(len(data) - lookback):

        # log.debug (f"data my_generator_candle {data} lookback {lookback}")

        temp_list = []
        for candle in data[first_row : first_row + lookback]:

            converted_data = ohlc_to_candlestick(candle)
            # log.info (f"converted_data  {converted_data} candle {candle}")
            temp_list.append(converted_data)
            # log.info (f"temp_list  {temp_list}")

        # log.error (f"temp_list  {temp_list}")
        temp_list2 = np.asarray(temp_list)
        templist3 = [temp_list2]
        templist4 = np.asarray(templist3, dtype="f4")
        # log.info (f"templist4  {templist4}")
        # log.warning (f"arr  1 {arr}")
        arr = np.append(arr, templist4, axis=0)
        # log.warning (f"arr  2 {arr}")
        first_row = first_row + 1

    return arr


def candles_analysis(
    np: object,
    ohlc_without_ticks: list,
    dim_sequence: int = 3,
):
    """
    https://www.tradingview.com/script/uuinZwsR-Big-Bar-Strategy/

    """

    dtype = [
        ("open", "f4"),
        ("high", "f4"),
        ("low", "f4"),
        ("close", "f4"),
    ]

    np_users_data = np.array(ohlc_without_ticks)

    np_data = np.array(
        [tuple(user.values()) for user in np_users_data],
        dtype=dtype,
    )

    candles_arrays = my_generator_candle(
        np,
        np_data[1:],
        dim_sequence,
    )

    candle_type = candles_arrays[-1, :, 0]  # (last_column_third_row)
    wicks_up = candles_arrays[-1, :, 1]  # (last_column_third_row)
    wicks_down = candles_arrays[-1, :, 2]  # (last_column_third_row)
    body_size = candles_arrays[-1, :, 3]  # (last_column_third_row)
    body_length = candles_arrays[-1, :, 4]  # (last_column_third_row)
    is_long_body = candles_arrays[-1, :, 5]  # (last_column_third_row)
    avg_body_length = np.average(body_length)
    body_length_exceed_average = body_length > avg_body_length
    # print(candles_arrays)
    # log.warning (f"candles_arrays {candles_arrays}")
    # log.error (f"candle_type {candle_type}")
    # log.warning (f"wicks_up {wicks_up}")
    # log.warning (f"wicks_down {wicks_down}")
    # log.warning (f"body_size {body_size}")
    # log.warning (f"body_length {body_length}")
    # log.warning (f"is_long_body {is_long_body}")
    # log.warning (f" avg_body_length {avg_body_length}")
    # log.warning (f" body_length_exceed_average {body_length_exceed_average}")

    return dict(
        candle_type=candle_type,
        body_length_exceed_average=body_length_exceed_average,
        is_long_body=(is_long_body),
    )


async def get_and_clean_up_candles_data(
    currency: str,
    resolution: int,
    qty_candles: int,
):
    """ """

    table_ohlc = f"ohlc{resolution}_{currency.lower()}_perp_json"

    ohlc_query = f"SELECT data FROM {table_ohlc} ORDER BY tick DESC LIMIT {qty_candles}"

    result_from_sqlite = await db_mgt.executing_query_with_return(ohlc_query)

    ohlc_without_volume = str_mod.remove_list_elements(
        str_mod.remove_apostrophes_from_json(o["data"] for o in result_from_sqlite),
        "volume",
    )

    ohlc_without_cost = str_mod.remove_list_elements(
        ohlc_without_volume,
        "cost",
    )

    ohlc_without_ticks = str_mod.remove_list_elements(
        ohlc_without_cost,
        "tick",
    )

    return ohlc_without_ticks


async def combining_candles_data(
    np: object,
    currencies: list,
    resolutions: int,
    qty_candles: int,
    dim_sequence: int = 3,
):
    """ """

    result = []
    for currency in currencies:

        instrument_name = f"{currency}-PERPETUAL"
        analysis_result = []
        for resolution in resolutions:

            if resolution != 1:

                candles_per_resolution = await get_and_clean_up_candles_data(
                    currency,
                    resolution,
                    qty_candles,
                )

                candles_analysis_result = candles_analysis(
                    np,
                    candles_per_resolution,
                    dim_sequence,
                )

                analysis_result.append(
                    dict(
                        resolution=(resolution),
                        candles_analysis=(candles_analysis_result),
                    )
                )

        result.append(
            dict(
                instrument_name=instrument_name,
                result=(analysis_result),
            )
        )

    return result


def translate_candles_data_to_market_condition(
    np: object,
    candles_data_instrument: list,
) -> dict:
    """ """
    try:

        # log.info(f"translate_candles_data_to_market_condition {candles_data_instrument}")

        candle_60 = [
            o["candles_analysis"]
            for o in candles_data_instrument
            if o["resolution"] == 60
        ]

        candle_60_type = np.sum([o["candle_type"] for o in candle_60])

        candle_60_is_long = np.sum([o["is_long_body"] for o in candle_60])

        candle_5 = [
            o["candles_analysis"]
            for o in candles_data_instrument
            if o["resolution"] == 5
        ]

        candle_5_type = np.sum([o["candle_type"] for o in candle_5])

        candle_5_is_long = np.sum([o["is_long_body"] for o in candle_5])

        candle_15 = [
            o["candles_analysis"]
            for o in candles_data_instrument
            if o["resolution"] == 15
        ]

        candle_15_type = np.sum([o["candle_type"] for o in candle_15])

        candle_15_is_long = np.sum([o["is_long_body"] for o in candle_15])

        candle_60_long_body_more_than_2 = candle_60_is_long >= 2
        candle_5_long_body_any = candle_5_is_long > 0
        candle_15_long_body_any = candle_15_is_long > 0
        candle_60_long_body_any = candle_60_is_long > 0

        candle_60_no_long = candle_60_is_long == 0

        neutral = True
        weak_bullish, weak_bearish = False, False
        bullish, bearish = False, False
        strong_bullish, strong_bearish = False, False

        if candle_5_long_body_any:
            weak_bullish = True if candle_5_type > 0 else False
            weak_bearish = True if candle_5_type < 0 else False

        if candle_60_long_body_any and candle_15_long_body_any:
            bullish = True if weak_bullish and candle_15_type > 0 else False
            bearish = True if weak_bearish and candle_15_type < 0 else False

        if candle_60_long_body_more_than_2:
            strong_bullish = (
                True if bullish and candle_60_long_body_more_than_2 else False
            )

            strong_bearish = (
                True if bearish and candle_60_long_body_more_than_2 else False
            )

        neutral = True if not weak_bearish and not weak_bullish else False

        return dict(
            strong_bullish=strong_bullish,
            bullish=bullish,
            weak_bullish=weak_bullish,
            neutral=neutral,
            weak_bearish=weak_bearish,
            bearish=bearish,
            strong_bearish=strong_bearish,
        )

    except Exception as error:

        error_handling.parse_error_message(error)


async def get_market_condition(
    client_redis: object,
    config_app: list,
    currencies,
    redis_channels,
    resolutions,
    np: object,
) -> dict:
    """ """
    try:

        # connecting to redis pubsub
        pubsub: object = client_redis.pubsub()

        market_analytics_channel: str = redis_channels["market_analytics_update"]
        chart_low_high_tick_channel: str = redis_channels["chart_low_high_tick"]

        redis_keys: dict = config_app["redis_keys"][0]
        market_condition_keys: str = redis_keys["market_condition"]

        # prepare channels placeholders
        channels = [
            #            market_analytics_channel,
            chart_low_high_tick_channel,
        ]

        # subscribe to channels
        [await pubsub.subscribe(o) for o in channels]

        qty_candles = 5
        dim_sequence = 3

        candles_data = await combining_candles_data(
            np,
            currencies,
            resolutions,
            qty_candles,
            dim_sequence,
        )

        # log.debug(f"candles_data {candles_data}")

        market_analytics_data = []

        while True:

            try:
                message_byte = await pubsub.get_message()

                if message_byte and message_byte["type"] == "message":

                    message_byte_data = orjson.loads(message_byte["data"])

                    params = message_byte_data["params"]

                    data = params["data"]

                    message_channel = params["channel"]

                    if chart_low_high_tick_channel in message_channel:

                        message_byte_data["params"].update(
                            {"channel": market_analytics_channel}
                        )

                        if market_analytics_data:

                            instrument_name = data["instrument_name"]

                            if instrument_name in data["instrument_name"]:

                                resolution = data["resolution"]

                                currency = data["currency"]

                                if resolution != 1:

                                    candles_per_resolution = (
                                        await get_and_clean_up_candles_data(
                                            currency,
                                            resolution,
                                            qty_candles,
                                        )
                                    )

                                    candles_data_instrument = candles_analysis(
                                        np,
                                        candles_per_resolution,
                                        dim_sequence,
                                    )

                                    # log.info(f"candles_data_instrument {candles_data_instrument}")

                                    candles_data_instrument = [
                                        o["result"]
                                        for o in candles_data
                                        if instrument_name in o["instrument_name"]
                                    ][0]

                                    pub_message = (
                                        translate_candles_data_to_market_condition(
                                            np,
                                            candles_data_instrument,
                                        )
                                    )

                                    market_analytics_data = [
                                        o
                                        for o in market_analytics_data
                                        if instrument_name not in o["instrument_name"]
                                    ]

                                    pub_message.update(
                                        {"instrument_name": instrument_name}
                                    )

                                    market_analytics_data.append(pub_message)

                                    message_byte_data["params"].update(
                                        {"data": market_analytics_data}
                                    )

                                    await redis_client.saving_and_publishing_result(
                                        client_redis,
                                        market_analytics_channel,
                                        market_condition_keys,
                                        message_byte_data,
                                        message_byte_data,
                                    )

                                    # log.debug(f"market_analytics_data {market_analytics_data}")

                        else:

                            candles_instrument_name = str_mod.remove_redundant_elements(
                                [o["instrument_name"] for o in candles_data]
                            )

                            for instrument_name in candles_instrument_name:

                                candles_data_instrument = [
                                    o["result"]
                                    for o in candles_data
                                    if instrument_name in o["instrument_name"]
                                ][0]

                                pub_message = (
                                    translate_candles_data_to_market_condition(
                                        np,
                                        candles_data_instrument,
                                    )
                                )

                                pub_message.update({"instrument_name": instrument_name})

                                market_analytics_data.append(pub_message)

                            message_byte_data["params"].update(
                                {"data": market_analytics_data}
                            )

                            await redis_client.saving_and_publishing_result(
                                client_redis,
                                market_analytics_channel,
                                market_condition_keys,
                                message_byte_data,
                                message_byte_data,
                            )

            except Exception as error:

                error_handling.parse_error_message(error)

                continue

            finally:
                await asyncio.sleep(0.1)

    except Exception as error:

        error_handling.parse_error_message(error)



# -*- coding: utf-8 -*-

# built ins
import os
import pickle


def read_data(file_name_pkl: str) -> list:
    """ """

    try:
        if os.path.exists(file_name_pkl):
            with open(file_name_pkl, "rb") as handle:
                read_pickle = pickle.load(handle)
                return read_pickle
    except:
        return []


def check_duplicate_elements(file_name: str) -> None:
    from utilities import string_modification

    # from loguru import logger as log

    data_from_db: list = read_data(file_name)

    free_from_duplicates_data = string_modification.remove_redundant_elements(
        data_from_db
    )

    dump_data_as_list(file_name, free_from_duplicates_data)


def dump_data_as_list(
    file_name: str, data: dict, check_duplicates: bool = False
) -> None:
    """ """

    with open(file_name, "wb") as handle:
        try:
            # print (f'dump_data_as_list {data}')

            if data != []:
                #    print (f'isinstance(data, dict) {isinstance(data, dict)}')
                #    print (f'isinstance(data, list) {isinstance(data, list)}')

                if isinstance(data, dict):
                    pickle.dump([data], handle, protocol=pickle.HIGHEST_PROTOCOL)

                if isinstance(data, list):
                    free_from_none_data = [o for o in data if isinstance(o, dict)]

                    pickle.dump(
                        free_from_none_data,
                        handle,
                        protocol=pickle.HIGHEST_PROTOCOL,
                    )

            if data == []:
                data_from_db: list = read_data(file_name)

                # to avoid record [] in db with valid contents
                if data_from_db == []:
                    pickle.dump(data, handle, protocol=pickle.HIGHEST_PROTOCOL)

                if data_from_db == None:
                    pickle.dump(data, handle, protocol=pickle.HIGHEST_PROTOCOL)

        except Exception as error:
            print(f"pickling {error}")

    if check_duplicates == True:
        check_duplicate_elements(file_name)


def append_data(file_name_pkl: str, data: dict, check_duplicates: bool = False) -> None:
    """
    https://stackoverflow.com/questions/28077573/python-appending-to-a-pickled-list
    """

    data_from_db = []
    # print (f"append_data data fr exc {data}")

    # collected_data: list = []
    if os.path.exists(file_name_pkl):
        data_from_db = read_data(file_name_pkl)

    if isinstance(data, list) and data != []:
        data = data[0]

    if data_from_db != []:
        data_from_db.append(data)

    combined_data = [data] if data_from_db == [] else data_from_db

    # Now we "sync" our database
    dump_data_as_list(file_name_pkl, combined_data, check_duplicates)


def replace_data(file_name: str, data: dict, check_duplicates: bool = False) -> None:
    """ """

    read = read_data(file_name)

    if read == []:
        pass

    dump_data_as_list(file_name, data, check_duplicates)


def append_and_replace_items(
    file_name_pkl: str, data: dict, check_duplicates: bool = False
) -> None:
    """
    append_and_replace_items (file_name, resp, 3)
    """

    append_data(file_name_pkl, data)

    data_from_db: list = read_data(file_name_pkl)
    # print (f"pickle isinstance(data_from_db, dict) { isinstance(data_from_db, dict)}")

    if isinstance(data_from_db, dict):
        data_list = list(data_from_db)

    if isinstance(data_from_db, list):
        try:
            data_list = list(data_from_db[0])
        except:
            print(f"except because of dict type. WEIRD. RECHECK) {data_from_db}")
            data_list = list(data_from_db)

    dump_data_as_list(file_name_pkl, data_list, check_duplicates)


def append_and_replace_items_based_on_qty(
    file_name_pkl: str,
    data: dict,
    max_qty: int,
    check_duplicates: bool = False,
) -> None:
    """
    append_and_replace_items_based_on_qty (file_name, resp, 3)
    """

    # print (f"pickle data fr exc {data}")

    append_data(file_name_pkl, data)
    data_from_db: list = read_data(file_name_pkl)

    # print (f"pickle isinstance(data_from_db, dict) { isinstance(data_from_db, dict)}")

    if isinstance(data_from_db, dict):
        data_list = list(data_from_db)

    if isinstance(data_from_db, list):
        try:
            data_list = list(data_from_db[0])
        except:
            print(f"except because of dict type. WEIRD. RECHECK) {data_from_db}")
            data_list = list(data_from_db)

    # print (f"append_and_replace_items_based_on_qty {data_from_db}")
    # data_list = list (data [0])

    if "change_id" in data_list:
        sorted_data: list = sorted([o["timestamp"] for o in data_from_db])
        len_tick_data_ordBook: int = len(sorted_data)

        if len_tick_data_ordBook > max_qty:
            filtered_timestamps = (sorted_data)[max_qty:]

            result: list = [
                o for o in data_from_db if o["timestamp"] not in filtered_timestamps
            ]

            dump_data_as_list(file_name_pkl, result, check_duplicates)

    if "params" in data_list:
        data_from_db = [o["params"] for o in data_from_db]

        data_from_db = [o["data"] for o in data_from_db]

        sorted_data: list = sorted([o["tick"] for o in data_from_db])
        len_tick_data_ohlc: int = len([o["tick"] for o in sorted_data])

        if len_tick_data_ohlc > max_qty:
            filtered_timestamps = (sorted_data)[max_qty:]

            result: list = [
                o for o in data_from_db if o["tick"] not in filtered_timestamps
            ]

            dump_data_as_list(file_name_pkl, result, check_duplicates)


def append_and_replace_items_based_on_time_expiration(
    file_name_pkl: str,
    data: dict,
    time_expiration: int,
    check_duplicates: bool = False,
) -> None:
    """
    append_and_replace_items_based_on_time_expiration in minute
    """
    from utilities import time_modification

    append_data(file_name_pkl, data)
    data: object = read_data(file_name_pkl)

    if isinstance(data, dict):
        data_list = list(data)

    if isinstance(data, list):
        data_list = list(data[0])

    # data_list = list (data [0])
    now_time_utc = time_modification.convert_time_to_utc()["utc_now"]
    now_time_utc_in_unix = time_modification.convert_time_to_unix(now_time_utc)

    one_hour_ago = now_time_utc_in_unix - time_expiration

    if "change_id" in data_list:
        result: list = [o for o in data if o["timestamp"] > one_hour_ago]

        dump_data_as_list(file_name_pkl, result, check_duplicates)

    if "params" in data_list:
        data = [o["params"] for o in data]

        data = [o["data"] for o in data]

        result: list = [o for o in data if o["tick"] < one_hour_ago]

        dump_data_as_list(file_name_pkl, result)

    if "with_rebates" in data_list and False:
        result: list = [
            o for o in data if o["remaining_active_time_in_hours"] == one_hour_ago
        ]
        dump_data_as_list(file_name_pkl, result, check_duplicates)



# -*- coding: utf-8 -*-
"""_summary_"""
# built ins
import asyncio

# installed
from loguru import logger as log

# user defined formula
from streaming_helper.db_management import sqlite_management as db_mgt
from streaming_helper.utilities import string_modification as str_mod


def get_settlement_period(strategy_attributes) -> list:

    return str_mod.remove_redundant_elements(
        str_mod.remove_double_brackets_in_list(
            [o["settlement_period"] for o in strategy_attributes]
        )
    )


def is_instrument_name_has_delivered(
    instrument_name: list, instrument_attributes_futures_all
) -> bool:

    active_futures_instrument = [
        o["instrument_name"] for o in instrument_attributes_futures_all
    ]

    return instrument_name not in active_futures_instrument


async def clean_up_closed_futures_because_has_delivered_(
    instrument_name, transaction, delivered_transaction
) -> None:

    log.warning(f"instrument_name {instrument_name}")
    log.warning(f"transaction {transaction}")
    try:
        trade_id_sqlite = int(transaction["trade_id"])

    except:
        trade_id_sqlite = transaction["trade_id"]

    timestamp = transaction["timestamp"]

    closed_label = f"futureSpread-closed-{timestamp}"

    transaction.update({"instrument_name": instrument_name})
    transaction.update({"timestamp": timestamp})
    transaction.update({"price": transaction["price"]})
    transaction.update({"amount": transaction["amount"]})
    transaction.update({"label": transaction["label"]})
    transaction.update({"trade_id": trade_id_sqlite})
    transaction.update({"order_id": transaction["order_id"]})

    # log.warning(f"transaction {transaction}")
    await db_mgt.insert_tables("my_trades_closed_json", transaction)

    await db_mgt.deleting_row(
        "my_trades_all_json",
        "databases/trading.sqlite3",
        "trade_id",
        "=",
        trade_id_sqlite,
    )

    delivered_transaction = delivered_transaction[0]

    timestamp_from_transaction_log = delivered_transaction["timestamp"]

    try:
        price_from_transaction_log = delivered_transaction["price"]

    except:
        price_from_transaction_log = delivered_transaction["index_price"]

    closing_transaction = transaction
    closing_transaction.update({"label": closed_label})
    closing_transaction.update({"amount": (closing_transaction["amount"]) * -1})
    closing_transaction.update({"price": price_from_transaction_log})
    closing_transaction.update({"trade_id": None})
    closing_transaction.update({"order_id": None})
    closing_transaction.update({"timestamp": timestamp_from_transaction_log})

    await db_mgt.insert_tables("my_trades_closed_json", closing_transaction)


async def updating_delivered_instruments(
    archive_db_table: str, instrument_name: str
) -> None:
    """ """

    where_filter = f"instrument_name"

    await db_mgt.update_status_data(
        archive_db_table, "is_open", where_filter, instrument_name, 0, "="
    )



# -*- coding: utf-8 -*-

# built ins
import asyncio

# installed
from loguru import logger as log

# user defined formulas
from streaming_helper.restful_api.deribit import end_point_params_template
from streaming_helper.channel_management import get_published_messages
from streaming_helper.channel_management.deribit import subscribing_to_channels
from streaming_helper.data_cleaning import managing_closed_transactions, reconciling_db
from streaming_helper.db_management import redis_client, sqlite_management as db_mgt
from streaming_helper.data_announcer.deribit import starter
from streaming_helper.utilities import (
    string_modification as str_mod,
    error_handling,
    system_tools,
    time_modification as time_mod,
)


async def reconciling_size(
    client_id: str,
    client_secret: str,
    client_redis: object,
    redis_channels: list,
    config_app: list,
    initial_data_subaccount: dict,
    futures_instruments: list,
) -> None:

    try:

        # connecting to redis pubsub
        pubsub: object = client_redis.pubsub()

        #instantiate api for private connection
        api_request: object = end_point_params_template.SendApiRequest(client_id,client_secret)

        # get tradable strategies
        tradable_config_app = config_app["tradable"]

        # get TRADABLE currencies
        currencies: list = [o["spot"] for o in tradable_config_app][0]

        # get redis channels
        order_allowed_channel: str = redis_channels["order_is_allowed"]
        positions_update_channel: str = redis_channels["position_cache_updating"]
        ticker_cached_channel: str = redis_channels["ticker_cache_updating"]
        sub_account_cached_channel: str = redis_channels["sub_account_cache_updating"]
        my_trade_receiving_channel: str = redis_channels["my_trade_receiving"]
        portfolio_channel: str = redis_channels["portfolio"]

        # subscribe to channels
        await subscribing_to_channels.redis_channels(
            pubsub,
            redis_channels,
            "reconciling_size",
        )

        server_time = time_mod.get_now_unix_time()

        ONE_SECOND = 1000

        one_minute = ONE_SECOND * 60

        min_expiration_timestamp = futures_instruments["min_expiration_timestamp"]

        active_futures = futures_instruments["active_futures"]

        all_instruments_name = futures_instruments["instruments_name"]

        futures_instruments_name = [o for o in all_instruments_name if "-FS-" not in o]

        result_template = str_mod.message_template()

        result = str_mod.message_template()
        initial_data_order_allowed = starter.is_order_allowed_combining(
            all_instruments_name,
            order_allowed_channel,
            result_template,
        )

        combined_order_allowed = initial_data_order_allowed["params"]["data"]

        sub_account_cached = initial_data_subaccount["params"]["data"]

        positions_cached = sub_account_cached["positions_cached"]
        
        await redis_client.publishing_result(
            client_redis,
            initial_data_order_allowed,
        )

        query_variables = f"instrument_name, label, amount_dir as amount, trade_id, timestamp"
                
        while True:

            try:

                message_byte = await pubsub.get_message()

                params = await get_published_messages.get_redis_message(message_byte)

                data, message_channel = params["data"], params["channel"]

                if order_allowed_channel in message_channel:

                    not_allowed_instruments = [
                        o["instrument_name"]
                        for o in combined_order_allowed
                        if o["size_is_reconciled"] == 0
                    ]
                    
                    log.info(f"not_allowed_instruments {not_allowed_instruments}")

                    if not_allowed_instruments:
                        
                        for instrument_name in not_allowed_instruments:
                            
                            currency_lower = str_mod.extract_currency_from_text(instrument_name).lower()
                            
                            archive_db_table = f"my_trades_all_{currency_lower}_json"
                            
                            query_trades_basic = f"SELECT {query_variables}  FROM  {archive_db_table}"

                            query_trades_active_where = (
                                f"WHERE instrument_name LIKE '%{instrument_name}%' AND is_open = 1 ORDER BY timestamp DESC"
                            )

                            query_trades_active_currency = (
                                f"{query_trades_basic} {query_trades_active_where}"
                            )

                            my_trades_instrument_name = await db_mgt.executing_query_with_return(
                                query_trades_active_currency
                            )
                            
                            my_trades_and_sub_account_size_reconciled = (
                                reconciling_db.is_my_trades_and_sub_account_size_reconciled_each_other(
                                    instrument_name,
                                    my_trades_instrument_name,
                                    positions_cached,
                                )
                            )

                            log.critical(
                                f"{instrument_name} {my_trades_and_sub_account_size_reconciled}"
                            )
                            
                            if not my_trades_and_sub_account_size_reconciled:                            

                                query_trades_where = (
                                    f"WHERE instrument_name LIKE '%{instrument_name}%' ORDER BY timestamp DESC LIMIT 2"
                                )

                                query_trades_instrument_name = (
                                    f"{query_trades_basic} {query_trades_where}"
                                )

                                my_trades_instrument_name = await db_mgt.executing_query_with_return(
                                    query_trades_instrument_name
                                )
                                
                                log.info(f" my_trades_instrument_name {my_trades_instrument_name}")
                                
                                if my_trades_instrument_name:
                                        
                                    min_timestamp = min([o["timestamp"] for o in my_trades_instrument_name])
                                
                                else:
                                    
                                    one_day_ago = server_time - (one_minute * 60 * 24 * 100)
                                        
                                    min_timestamp = one_day_ago
                                
                                transaction_log = await api_request.get_transaction_log(
                                        currency_lower,
                                        min_timestamp,
                                        1000,
                                        "trade",)
                                
                                if transaction_log:
                                    
                                    await distributing_transaction_log_from_exchange(
                                        archive_db_table,
                                        instrument_name,
                                        transaction_log,
                                        )

                if ticker_cached_channel in message_channel:

                    exchange_server_time = data["server_time"]

                    delta_time = (exchange_server_time - server_time) / ONE_SECOND

                    if delta_time > 5:

                        await rechecking_reconciliation_regularly(
                            client_redis,
                            combined_order_allowed,
                            futures_instruments_name,
                            currencies,
                            order_allowed_channel,
                            positions_cached,
                            result,
                        )

                        server_time = exchange_server_time

                if (
                    positions_update_channel in message_channel
                    or sub_account_cached_channel in message_channel
                    or my_trade_receiving_channel in message_channel
                    or portfolio_channel in message_channel
                ):

                    for currency in currencies:

                        currency_lower = currency.lower()

                        archive_db_table = f"my_trades_all_{currency_lower}_json"

                        await inserting_transaction_log_data(
                            api_request,
                            archive_db_table,
                            currency,
                        )

                    if sub_account_cached_channel in message_channel:
                        positions_cached = data["positions"]

                    else:
                        try:
                            positions_cached = data["positions"]

                        except:

                            positions_cached = positions_cached

                    await rechecking_reconciliation_regularly(
                        client_redis,
                        combined_order_allowed,
                        futures_instruments_name,
                        currencies,
                        order_allowed_channel,
                        positions_cached,
                        result,
                    )

            except Exception as error:

                await error_handling.parse_error_message_with_redis(
                    client_redis,
                    error,
                    )

                continue

            finally:
                await asyncio.sleep(0.001)

    except Exception as error:

        await error_handling.parse_error_message_with_redis(
            client_redis,
            error,
        )


async def rechecking_reconciliation_regularly(
    client_redis: object,
    combined_order_allowed: list,
    futures_instruments_name,
    currencies: list,
    order_allowed_channel: str,
    positions_cached: list,
    result,
) -> None:
    """ """

    positions_cached_all = str_mod.remove_redundant_elements(
        [o["instrument_name"] for o in positions_cached]
    )

    # eliminating combo transactions as they're not recorded in the book
    positions_cached_instrument = [o for o in positions_cached_all if "-FS-" not in o]

    futures_instruments_name_not_in_positions_cached_instrument = [
        list(set(futures_instruments_name).difference(positions_cached_instrument))
    ][0]

    await allowing_order_for_instrument_not_in_sub_account(
        client_redis,
        combined_order_allowed,
        order_allowed_channel,
        futures_instruments_name_not_in_positions_cached_instrument,
        result,
    )

    await rechecking_based_on_sub_account(
        client_redis,
        combined_order_allowed,
        order_allowed_channel,
        positions_cached,
        positions_cached_instrument,
        result,
    )

    await rechecking_based_on_data_in_sqlite(
        client_redis,
        combined_order_allowed,
        currencies,
        order_allowed_channel,
        positions_cached,
        result,
    )


async def allowing_order_for_instrument_not_in_sub_account(
    client_redis: object,
    combined_order_allowed: list,
    order_allowed_channel: str,
    futures_instruments_name_not_in_positions_cached_instrument: list,
    result: dict,
) -> None:
    """ """

    order_allowed = 1

    for instrument_name in futures_instruments_name_not_in_positions_cached_instrument:

        [o for o in combined_order_allowed if instrument_name in o["instrument_name"]][
            0
        ]["size_is_reconciled"] = order_allowed

    result["params"].update({"channel": order_allowed_channel})
    result["params"].update({"data": combined_order_allowed})

    await redis_client.publishing_result(
        client_redis,
        result,
    )


async def rechecking_based_on_sub_account(
    client_redis: object,
    combined_order_allowed: list,
    order_allowed_channel: str,
    positions_cached: list,
    positions_cached_instrument: list,
    result: dict,
) -> None:
    """ """

    # FROM sub account to other db's
    if positions_cached_instrument:

        # sub account instruments
        for instrument_name in positions_cached_instrument:

            currency: str = str_mod.extract_currency_from_text(instrument_name)

            currency_lower = currency.lower()

            archive_db_table = f"my_trades_all_{currency_lower}_json"

            query_trades = f"SELECT * FROM  v_{currency_lower}_trading_active"

            my_trades_currency_all_transactions: list = (
                await db_mgt.executing_query_with_return(query_trades)
            )

            # handling transactions with no label
            await labelling_blank_labels(
                instrument_name,
                my_trades_currency_all_transactions,
                archive_db_table,
            )

            my_trades_instrument_name = (
                []
                if my_trades_currency_all_transactions == []
                else [
                    o
                    for o in my_trades_currency_all_transactions
                    if instrument_name in o["instrument_name"]
                ]
            )

            await managing_closed_transactions.clean_up_closed_transactions(
                archive_db_table,
                my_trades_instrument_name,
            )

            my_trades_and_sub_account_size_reconciled = (
                reconciling_db.is_my_trades_and_sub_account_size_reconciled_each_other(
                    instrument_name,
                    my_trades_instrument_name,
                    positions_cached,
                )
            )

            log.critical(
                f"{instrument_name} {my_trades_and_sub_account_size_reconciled}"
            )

            updating_order_allowed_cache(
                combined_order_allowed,
                instrument_name,
                my_trades_and_sub_account_size_reconciled,
            )

        result["params"].update({"channel": order_allowed_channel})
        result["params"].update({"data": combined_order_allowed})

        await redis_client.publishing_result(
            client_redis,
            result,
        )


async def rechecking_based_on_data_in_sqlite(
    client_redis: object,
    combined_order_allowed: list,
    currencies: list,
    order_allowed_channel: str,
    positions_cached: list,
    result: dict,
) -> None:
    """ """

    for currency in currencies:

        currency_lower = currency.lower()

        archive_db_table = f"my_trades_all_{currency_lower}_json"
        
        query_variables = f"instrument_name, label, amount_dir as amount, timestamp, trade_id"

        query_trades_active_basic = f"SELECT {query_variables}  FROM  {archive_db_table}"

        query_trades_active_where = (
            f"WHERE instrument_name LIKE '%{currency}%' AND is_open = 1"
        )

        query_trades_active_currency = (
            f"{query_trades_active_basic} {query_trades_active_where}"
        )

        my_trades_active_currency = await db_mgt.executing_query_with_return(
            query_trades_active_currency
        )

        my_trades_active_instrument = str_mod.remove_redundant_elements(
            [o["instrument_name"] for o in my_trades_active_currency]
        )

        if my_trades_active_instrument:

            # sub account instruments
            for instrument_name in my_trades_active_instrument:

                my_trades_active = [
                    o
                    for o in my_trades_active_currency
                    if instrument_name in o["instrument_name"]
                ]

                if my_trades_active:

                    my_trades_and_sub_account_size_reconciled = reconciling_db.is_my_trades_and_sub_account_size_reconciled_each_other(
                        instrument_name,
                        my_trades_active,
                        positions_cached,
                    )

                    log.critical(
                        f"{instrument_name} {my_trades_and_sub_account_size_reconciled}"
                    )

                    updating_order_allowed_cache(
                        combined_order_allowed,
                        instrument_name,
                        my_trades_and_sub_account_size_reconciled,
                    )

                    await managing_closed_transactions.clean_up_closed_transactions(
                        archive_db_table,
                        my_trades_active,
                    )

    result["params"].update({"channel": order_allowed_channel})
    result["params"].update({"data": combined_order_allowed})

    await redis_client.publishing_result(
        client_redis,
        result,
    )


def updating_order_allowed_cache(
    combined_order_allowed: list,
    instrument_name: str,
    my_trades_and_sub_account_size_reconciled: bool,
) -> None:
    """ """

    if my_trades_and_sub_account_size_reconciled:

        order_allowed = 1

    else:

        order_allowed = 0

    instrument_exist = [o for o in combined_order_allowed if instrument_name in o["instrument_name"]]

    if instrument_exist:
        
        [o for o in combined_order_allowed if instrument_name in o["instrument_name"]][0][
        "size_is_reconciled"
    ] = order_allowed


async def inserting_transaction_log_data(
    api_request: object,
    archive_db_table: str,
    currency: str,
) -> None:
    """ """

    query_trades_active_basic = f"SELECT instrument_name, user_seq, timestamp, trade_id  FROM  {archive_db_table}"

    query_trades_active_where = f"WHERE instrument_name LIKE '%{currency}%'"

    query_trades = f"{query_trades_active_basic} {query_trades_active_where}"

    my_trades_currency = await db_mgt.executing_query_with_return(query_trades)

    if my_trades_currency:

        my_trades_currency_with_blanks_user_seq = [
            o["timestamp"] for o in my_trades_currency if o["user_seq"] is None
        ]

        if my_trades_currency_with_blanks_user_seq:

            min_timestamp = min(my_trades_currency_with_blanks_user_seq) - 100000

            transaction_log = await api_request.get_transaction_log(
                                            currency,
                                            min_timestamp,
                                            100,
                                            "trade",)

            where_filter = f"trade_id"

            log.debug(
                f"my_trades_currency_with_blanks_user_seq {my_trades_currency_with_blanks_user_seq}"
            )
            log.warning(f"transaction_log {transaction_log}")
            log.info([o for o in my_trades_currency if o["user_seq"] is None])
            
            if transaction_log:
                for transaction in transaction_log:

                    trade_id = transaction["trade_id"]
                    user_seq = int(transaction["user_seq"])
                    side = transaction["side"]
                    timestamp = int(transaction["timestamp"])
                    position = transaction["position"]
                    
                    components = ["user_seq", "side", "timestamp", "position"]
                    
                    for component in components:    
                        
                        if component == "user_seq":
                            sub_component = user_seq
                        
                        elif component == "side":
                            sub_component = side
                        
                        elif component == "timestamp":
                            sub_component = timestamp
                        
                        elif component == "position":
                            sub_component = position
                        
                        await db_mgt.update_status_data(
                            archive_db_table, 
                            component, 
                            where_filter,
                            trade_id, 
                            sub_component, 
                            "=",
                            )
                        

async def labelling_blank_labels(
    instrument_name: str,
    my_trades_currency_active: list,
    archive_db_table: str,
) -> None:

    my_trades_currency_active_with_blanks = [
        o for o in my_trades_currency_active if o["label"] is None
    ]

    log.debug(
        f"my_trades_currency_active_with_blanks {my_trades_currency_active_with_blanks}"
    )

    if my_trades_currency_active_with_blanks:
        column_trade: str = (
            "id",
            "instrument_name",
            "data",
            "label",
            "trade_id",
        )

        my_trades_currency_archive: list = (
            await db_mgt.executing_query_based_on_currency_or_instrument_and_strategy(
                archive_db_table, instrument_name, "all", "all", column_trade
            )
        )

        my_trades_currency_active_with_blanks = [
            o for o in my_trades_currency_archive if o["label"] is None
        ]

        my_trades_archive_instrument_id = [
            o["trade_id"] for o in my_trades_currency_active_with_blanks
        ]

        if my_trades_archive_instrument_id:
            for id in my_trades_archive_instrument_id:

                transaction = str_mod.parsing_sqlite_json_output(
                    [
                        o["data"]
                        for o in my_trades_currency_active_with_blanks
                        if id == o["trade_id"]
                    ]
                )[0]

                log.warning(f"transaction {transaction}")

                label_open: str = get_custom_label(transaction)

                where_filter = "trade_id"

                await db_mgt.update_status_data(
                    archive_db_table,
                    "label",
                    where_filter,
                    id,
                    label_open,
                    "=",
                )


def get_custom_label(transaction: list) -> str:

    side = transaction["direction"]
    side_label = "Short" if side == "sell" else "Long"

    try:
        last_update = transaction["timestamp"]
    except:
        try:
            last_update = transaction["last_update_timestamp"]
        except:
            last_update = transaction["creation_timestamp"]

    return f"custom{side_label.title()}-open-{last_update}"


async def distributing_transaction_log_from_exchange(
    archive_db_table: str,
    instrument_name: str,
    transaction_log: list,
) -> None:

    if transaction_log and "too_many_requests" not in transaction_log:

        for transaction in transaction_log:
            
            transaction_instrument_name = transaction["instrument_name"]
            
            transaction_currency = transaction["currency"]

            transaction_currency_usd = (f"{transaction_currency.upper()}_USD")
            
            log.info(f"transaction_currency_usd {transaction_currency_usd} {transaction_currency_usd not in transaction_instrument_name}")
            
            if (
                instrument_name in transaction_instrument_name 
                and transaction_currency_usd not in transaction_instrument_name):
                
                result = {}

                if "sell" in transaction["side"]:
                    direction = "sell"

                if "buy" in transaction["side"]:
                    direction = "buy"

                result.update({"trade_id": transaction["trade_id"]})
                result.update({"user_seq": transaction["user_seq"]})
                result.update({"side": transaction["side"]})
                result.update({"timestamp": transaction["timestamp"]})
                result.update({"position": transaction["position"]})
                result.update({"amount": transaction["amount"]})
                result.update({"order_id": transaction["order_id"]})
                result.update({"price": transaction["price"]})
                result.update({"instrument_name": transaction_instrument_name})
                result.update({"label": None})
                result.update({"direction": direction})
                result.update({"currency": transaction_currency})

                await db_mgt.insert_tables(
                    archive_db_table,
                    result,
                )





# -*- coding: utf-8 -*-

# built ins
import asyncio


# user defined formula
from streaming_helper.restful_api.deribit import end_point_params_template as end_point
from streaming_helper.utilities.pickling import read_data
from streaming_helper.utilities.string_modification import (
    remove_double_brackets_in_list,
)
from streaming_helper.utilities.system_tools import provide_path_for_file


def get_instruments_kind(
    currency: str,
    settlement_periods: list,
    kind: str = "all",
    result: list = None,
) -> list:
    """_summary_

    Args:
        currency (str): _description_
        kind (str): "future_combo",  "future"
        Instance:  [
                    {'tick_size_steps': [], 'quote_currency': 'USD', 'min_trade_amount': 1,'counter_currency': 'USD',
                    'settlement_period': 'month', 'settlement_currency': 'ETH', 'creation_timestamp': 1719564006000,
                    'instrument_id': 342036, 'base_currency': 'ETH', 'tick_size': 0.05, 'contract_size': 1, 'is_active': True,
                    'expiration_timestamp': 1725004800000, 'instrument_type': 'reversed', 'taker_commission': 0.0,
                    'maker_commission': 0.0, 'instrument_name': 'ETH-FS-27SEP24_30AUG24', 'kind': 'future_combo',
                    'rfq': False, 'price_index': 'eth_usd'}, ]
     Returns:
        list: _description_


    """

    if result:

        result = result["result"]

    else:
        my_path_instruments = provide_path_for_file("instruments", currency)

        instruments_raw = read_data(my_path_instruments)

        result = instruments_raw[0]["result"]

    non_spot_instruments = [o for o in result if o["kind"] != "spot"]
    instruments_kind = (
        non_spot_instruments
        if kind == "all"
        else [o for o in result if o["kind"] == kind]
    )

    return [o for o in instruments_kind if o["settlement_period"] in settlement_periods]


async def get_futures_for_active_currencies(
    active_currencies: list, 
    settlement_periods: list,
) -> list:
    """_summary_

    Returns:
        list: _description_
    """

    instruments_holder_place = []
    for currency in active_currencies:

        result = await end_point.get_instruments(currency)

        future_instruments = get_instruments_kind(
            currency, settlement_periods, "future", result
        )

        future_combo_instruments = get_instruments_kind(
            currency, settlement_periods, "future_combo", result
        )

        active_combo_perp = [
            o for o in future_combo_instruments if "_PERP" in o["instrument_name"]
        ]

        combined_instruments = future_instruments + active_combo_perp
        instruments_holder_place.append(combined_instruments)

    # removing inner list
    # typical result: [['BTC-30AUG24', 'BTC-6SEP24', 'BTC-27SEP24', 'BTC-27DEC24',
    # 'BTC-28MAR25', 'BTC-27JUN25', 'BTC-PERPETUAL'], ['ETH-30AUG24', 'ETH-6SEP24',
    # 'ETH-27SEP24', 'ETH-27DEC24', 'ETH-28MAR25', 'ETH-27JUN25', 'ETH-PERPETUAL']]

    instruments_holder_plc = []
    for instr in instruments_holder_place:
        instruments_holder_plc.append(instr)

    return remove_double_brackets_in_list(instruments_holder_plc)


async def get_futures_instruments(
    active_currencies: list,
    settlement_periods: list,
) -> dict:

    active_futures = await get_futures_for_active_currencies(
        active_currencies, settlement_periods
    )

    min_expiration_timestamp = min([o["expiration_timestamp"] for o in active_futures])

    return dict(
        instruments_name=[o["instrument_name"] for o in (active_futures)],
        min_expiration_timestamp=min_expiration_timestamp,
        active_futures=[o for o in active_futures if "future" in o["kind"]],
        active_combo=[o for o in active_futures if "future_combo" in o["kind"]],
        instruments_name_with_min_expiration_timestamp=[
            o["instrument_name"]
            for o in active_futures
            if o["expiration_timestamp"] == min_expiration_timestamp
        ][0],
    )



# # -*- coding: utf-8 -*-

# built ins
import asyncio

# installed
from dataclassy import dataclass, fields
from loguru import logger as log

from streaming_helper.strategies.deribit import basic_strategy as basic
from streaming_helper.strategies.deribit.basic_strategy import (
    get_label,
    get_label_integer,
)

from streaming_helper.utilities import (
    error_handling,
    pickling,
    string_modification as str_mod,
    system_tools,
)

# user defined formula


def reading_from_pkl_data(end_point: str, currency: str, status: str = None) -> dict:
    """ """

    path: str = system_tools.provide_path_for_file(end_point, currency, status)

    data = pickling.read_data(path)

    return data


def get_transactions_len(result_strategy_label) -> int:
    """ """
    return 0 if result_strategy_label == [] else len([o for o in result_strategy_label])


def get_transactions_len(result_strategy_label) -> int:
    """ """
    return 0 if result_strategy_label == [] else len([o for o in result_strategy_label])


def get_delta(my_trades_currency_strategy) -> int:
    """ """

    return (
        0
        if my_trades_currency_strategy == []
        else sum([o["amount"] for o in my_trades_currency_strategy])
    )


def convert_list_to_dict(transaction: list) -> dict:

    # convert list to dict
    try:
        transaction = transaction[0]
    except:
        return transaction

    return transaction


def get_basic_opening_size(
    notional: float,
    monthly_target_profit: float,
    average_movement: float,
    basic_ticks_for_average_meovement: int,
) -> int:
    """ """

    hour_in_minutes = 60

    target_profit_hourly = monthly_target_profit / 30 / 24

    usd_per_hour = notional * target_profit_hourly

    size_per_hour = usd_per_hour / average_movement

    return size_per_hour / (hour_in_minutes / basic_ticks_for_average_meovement)


def determine_opening_size(
    instrument_name: str,
    instrument_attributes_futures: list,
    notional: float,
    monthly_target_profit: float,
    average_movement: float,
    basic_ticks_for_average_meovement: int,
) -> int:
    """ """

    proposed_size = get_basic_opening_size(
        notional,
        monthly_target_profit,
        average_movement,
        basic_ticks_for_average_meovement,
    )

    return basic.size_rounding(
        instrument_name, instrument_attributes_futures, proposed_size
    )


def is_contango(
    price_future: float,
    price_perpetual: float,
) -> int:

    return price_future > price_perpetual


def determine_exit_side_combo_auto(
    traded_price_future: float,
    traded_price_perpetual: float,
    traded_side_future: str,
    traded_side_perpetual: str,
) -> str:
    """ """

    traded_transaction_is_contango = is_contango(
        traded_price_future, traded_price_perpetual
    )

    if traded_transaction_is_contango:

        if traded_side_future == "sell" and traded_side_perpetual == "buy":

            exit_side = "buy"

            return exit_side


def extracting_closing_parameters_from_transactions(
    traded_future: float,
    transactions_len: float,
    traded_side_future: str,
    traded_side_perpetual: str,
) -> str:
    """ """

    if transactions_len == 2:
        traded_future: dict = convert_list_to_dict(traded_future)
        price_future = traded_future["price"]
        instrument_name_future = traded_future["instrument_name"]

    else:
        price_future = traded_future["price"]
        instrument_name_future = traded_future["instrument_name"]

    return dict()


def get_label_main(result: list, strategy_label: str) -> list:
    """ """

    return [
        o
        for o in result
        if str_mod.parsing_label(strategy_label)["main"]
        == str_mod.parsing_label(o["label"])["main"]
    ]


def get_outstanding_closed_orders(orders_currency: dict, label_integer: int) -> list:
    """
    get outstanding closed orders for respective label
    """

    return [
        o
        for o in orders_currency
        if str(label_integer) in o["label"] and "closed" in o["label"]
    ]


def basic_ordering(orders_currency: dict, label_integer: int) -> list:
    """
    basic ordering
    """

    if orders_currency:
        outstanding_closed_orders = get_outstanding_closed_orders(
            orders_currency, label_integer
        )

    no_orders_at_all = not orders_currency

    current_order_not_related_to_respective_label = (
        orders_currency and not outstanding_closed_orders
    )

    return no_orders_at_all or current_order_not_related_to_respective_label


def get_transactions_premium(
    transactions: list,
) -> float:
    """ """

    return abs(
        sum([(o["price"] * o["amount"]) / abs(o["amount"]) for o in transactions])
    )


def creating_instrument_name_combo(
    traded_instrument_name_future: list,
) -> float:
    """ """

    return f"{traded_instrument_name_future[:3]}-FS-{traded_instrument_name_future[4:]}_PERP"


def delta_premium_pct(transactions_premium: float, current_premium: float) -> str:
    """ """

    return abs(current_premium - transactions_premium) / transactions_premium


def get_basic_opening_parameters(strategy_label):
    """ """

    # provide placeholder for params
    params = {}

    # default type: limit
    params.update({"type": "limit"})

    label_open: str = get_label("open", strategy_label)
    params.update({"label": label_open})

    return params


def check_if_minimum_waiting_time_has_passed(
    threshold: float,
    timestamp: int,
    server_time: int,
) -> bool:
    """ """

    cancel_allowed: bool = False

    minimum_waiting_time_has_passed: bool = basic.is_minimum_waiting_time_has_passed(
        server_time, timestamp, threshold
    )

    if minimum_waiting_time_has_passed:
        cancel_allowed: bool = True

    return cancel_allowed


def get_label(status: str, label_main_or_label_transactions: str) -> str:
    """
    provide transaction label
    """
    from utilities import label_numbering

    if status == "open":
        # get open label
        label = str_mod.labelling("open", label_main_or_label_transactions)

    if status == "closed":

        # parsing label id
        label_id: int = str_mod.parsing_label(label_main_or_label_transactions)["int"]

        # parsing label strategy
        label_main: str = str_mod.parsing_label(label_main_or_label_transactions)[
            "main"
        ]

        # combine id + label strategy
        label: str = f"""{label_main}-closed-{label_id}"""

    return label


def modified_tp_threshold(
    instrument_attributes_futures: list,
    take_profit_threshold_original: float,
    instrument_name,
) -> float:
    """ """

    instrument_attributes = [
        o
        for o in instrument_attributes_futures
        if instrument_name in o["instrument_name"]
    ][0]

    settlement_period = instrument_attributes["settlement_period"]

    return (
        take_profit_threshold_original * 2
        if settlement_period == "week"
        else take_profit_threshold_original * 5
    )  # 0.05%


def compare_transactions_price_against_threshold(
    transactions: list,
    current_price_future: float,
    side: str,
    tp_threshold: float,
) -> float:
    """ """
    return [
        o
        for o in transactions
        if current_price_future < (o["price"] - (o["price"] * tp_threshold))
        and o["side"] == side
    ]


def proforma_delta(delta: float, selected_transaction_size: int, side: str) -> float:

    selected_transaction = abs(selected_transaction_size)

    if side == "sell":
        selected_transaction = selected_transaction * -1

    return delta + selected_transaction


def is_contra_order_will_reduce_delta(
    delta: float, selected_transaction_size: int, side: str
) -> float:

    proforma = proforma_delta(delta, selected_transaction_size, side)

    if delta > 0:

        # log.warning (f"will_reduce_delta {proforma < delta} proforma {proforma} delta {delta}")

        return proforma < delta

    if delta < 0:
        # log.warning (f"will_reduce_delta {delta < proforma} proforma {proforma} delta {delta}")
        return delta < proforma


@dataclass(unsafe_hash=True, slots=True)
class ComboAuto(basic.BasicStrategy):
    """ """

    orders_currency_strategy: list
    server_time: int
    market_condition: list
    my_trades_currency_strategy: list
    ticker_perpetual: dict = None
    delta: float = fields
    basic_params: object = fields

    def __post_init__(self):

        self.delta: float = get_delta(self.my_trades_currency_strategy)
        self.basic_params: str = basic.BasicStrategy(
            self.strategy_label, self.strategy_parameters
        )

        # log.critical(f"""delta  {self.delta} """)

    async def is_send_open_order_allowed_auto_combo(
        self,
        ticker_future,
        ticker_combo,
        notional,
        instrument_name_combo,
        instrument_attributes_futures,
        instrument_attributes_combo,
        monthly_target_profit: float,
        average_movement: float,
        basic_ticks_for_average_meovement: int,
    ) -> dict:
        """ """
        # log.warning (f"auto_combo {instrument_name_combo}")

        order_allowed = False

        orders_currency = self.orders_currency_strategy

        orders_instrument: list = [
            o for o in orders_currency if instrument_name_combo in o["instrument_name"]
        ]

        open_orders_instrument: list = [
            o for o in orders_instrument if "open" in o["label"]
        ]

        if open_orders_instrument:

            len_open_orders_instrument: list = len(open_orders_instrument)

            try:
                last_order_time = max([o["timestamp"] for o in open_orders_instrument])

            except:
                last_order_time = max(
                    [o["last_update_timestamp"] for o in open_orders_instrument]
                )

            delta_time = self.server_time - last_order_time

            delta_time_seconds = delta_time / 1000

            threshold = 60 * 5

            max_stacked_orders = 1

        # log.error (f"open_orders_instrument {open_orders_instrument} ")
        if not open_orders_instrument or (
            delta_time_seconds > threshold
            and len_open_orders_instrument < max_stacked_orders
        ):

            ask_price_combo = ticker_combo["best_ask_price"]
            ask_price_future = ticker_future["best_ask_price"]
            bid_price_future = ticker_future["best_bid_price"]
            instrument_name_future = ticker_future["instrument_name"]
            ask_price_perpetual = self.ticker_perpetual["best_ask_price"]
            bid_price_perpetual = self.ticker_perpetual["best_bid_price"]

            contango = is_contango(
                ask_price_future,
                bid_price_perpetual,
            )

            if contango and ask_price_combo > 0:
                tick_size = instrument_attributes_combo["tick_size"]
                fair_value_combo = (
                    abs(bid_price_future - ask_price_perpetual) / tick_size
                ) * tick_size
                entry_price = max(fair_value_combo, ask_price_combo)

                order_allowed = True

        if order_allowed:

            size: int = determine_opening_size(
                instrument_name_future,
                instrument_attributes_futures,
                notional,
                monthly_target_profit,
                average_movement,
                basic_ticks_for_average_meovement,
            )

            label_open_minus_auto: str = get_label("open", self.strategy_label)

            label_integer: int = get_label_integer(label_open_minus_auto)

            label_open: str = f"{self.strategy_label}Auto-open-{label_integer}"

            # provide placeholder for params
            params = {}

            params.update({"instrument_name": instrument_name_combo})
            params.update({"side": "sell"})
            params.update({"size": size})
            params.update({"entry_price": entry_price})
            params.update({"label": label_open})

            # log.error (f"fair_value_combo {fair_value_combo} ask_price_combo {ask_price_combo} entry_price {entry_price} ")

            # default type: limit
            params.update({"type": "limit"})

        return dict(
            order_allowed=order_allowed,
            order_parameters=[] if order_allowed == False else params,
        )

    async def cancelling_orders(
        self,
        transaction: dict,
        server_time: int,
        strategy_params: list = None,
        orders_currency_strategy: list = None,
    ) -> bool:
        """ """
        cancel_allowed: bool = False

        if orders_currency_strategy is None:
            orders_currency = self.orders_currency_strategy

        label_integer = get_label_integer(transaction["label"])

        len_outstanding_closed_orders = len(
            [
                o["amount"]
                for o in orders_currency
                if str(label_integer) in o["label"] and "closed" in o["label"]
            ]
        )

        if len_outstanding_closed_orders > 1:
            cancel_allowed: bool = True

        else:
            ONE_SECOND = 1000
            ONE_MINUTE = ONE_SECOND * 60

            if strategy_params is None:
                strategy_params: dict = self.strategy_parameters

            waiting_minute_before_cancel = (
                strategy_params["waiting_minute_before_cancel"] * ONE_MINUTE
            )

            # log.warning(f"waiting_minute_before_cancel {waiting_minute_before_cancel} {waiting_minute_before_cancel {waiting_minute_before_cancel}}")
            # log.warning(f"transaction {transaction}")

            try:
                timestamp: int = transaction["timestamp"]
            except:
                timestamp: int = transaction["last_update_timestamp"]

            if "open" in transaction["label"]:

                cancel_allowed: bool = check_if_minimum_waiting_time_has_passed(
                    waiting_minute_before_cancel,
                    timestamp,
                    server_time,
                )

            if "closed" in transaction["label"]:

                cancel_allowed: bool = check_if_minimum_waiting_time_has_passed(
                    waiting_minute_before_cancel,
                    timestamp,
                    server_time,
                )

        # log.error (f"cancel_allowed {cancel_allowed} transaction {transaction} ")

        return cancel_allowed

    async def is_cancelling_orders_allowed(
        self,
        selected_transaction: list,
        server_time: int,
        strategy_params: list = None,
        orders_currency_strategy: list = None,
    ) -> dict:
        """ """

        cancel_allowed, cancel_id = False, None

        cancel_allowed = await self.cancelling_orders(
            selected_transaction,
            server_time,
            strategy_params,
            orders_currency_strategy,
        )

        if cancel_allowed:
            cancel_id = selected_transaction["order_id"]

        return dict(cancel_allowed=cancel_allowed, cancel_id=cancel_id)

    async def is_send_exit_order_allowed_combo_auto(
        self,
        label: str,
        instrument_attributes_combo_all: list,
        take_profit_threshold_original: float,
    ) -> dict:
        """
        Returns:
            dict: _description_
        """

        # log.info (f"is_send_exit_order_allowed_combo_auto {label}")

        order_allowed = False
        my_trades_currency = self.my_trades_currency_strategy

        if my_trades_currency:

            label_integer = get_label_integer(label)

            transactions = [
                o for o in my_trades_currency if str(label_integer) in o["label"]
            ]

            transactions_sum = sum([o["amount"] for o in transactions])

            transactions_len = len(
                transactions
            )  # sum product function applied only for 2 items.

            # log.info (f"transactions {transactions} ")
            # log.info (f"transactions_sum {transactions_sum} transactions_len {transactions_len}")

            if transactions_sum == 0 and transactions_len == 2:

                traded_future = [
                    o for o in transactions if "PERPETUAL" not in o["instrument_name"]
                ]

                if traded_future:

                    traded_future = traded_future[0]

                    traded_instrument_name_future = traded_future["instrument_name"]

                    instrument_name_combo = creating_instrument_name_combo(
                        traded_instrument_name_future
                    )

                    combo_ticker = reading_from_pkl_data(
                        "ticker", instrument_name_combo
                    )

                    if combo_ticker:

                        orders_currency = self.orders_currency_strategy

                        strategy_label = self.strategy_label

                        instrument_name_perpetual = self.ticker_perpetual[
                            "instrument_name"
                        ]

                        current_premium = combo_ticker[0]["best_bid_price"]

                        transactions_premium = get_transactions_premium(transactions)

                        premium_pct = delta_premium_pct(
                            transactions_premium,
                            current_premium,
                        )

                        basic_ordering_is_ok = basic_ordering(
                            orders_currency, label_integer
                        )

                        tp_threshold = modified_tp_threshold(
                            instrument_attributes_combo_all,
                            take_profit_threshold_original,
                            instrument_name_combo,
                        )

                        # log.debug  (f" {premium_pct > tp_threshold } basic_ordering_is_ok {basic_ordering_is_ok } current_premium > 0 {current_premium} {current_premium > 0} current_premium < transactions_premium {current_premium < transactions_premium}")

                        # log.warning (f"transactions {transactions}")
                        # log.warning (f"orders_currency {orders_currency}")
                        len_label = len(
                            [
                                o["label"]
                                for o in orders_currency
                                if str(label_integer) in o["label"]
                            ]
                        )

                        # log.debug (f"len_label {len_label} label_integer {label_integer}")

                        if (
                            len_label == 0
                            and premium_pct > tp_threshold
                            and basic_ordering_is_ok
                            and current_premium > 0
                            and current_premium < transactions_premium
                        ):
                            # log.error (f"transactions {transactions}")
                            traded_perpetual: list = [
                                o
                                for o in transactions
                                if instrument_name_perpetual in o["instrument_name"]
                            ][0]

                            traded_perpetual_size = abs(traded_perpetual["amount"])

                            traded_price_perpetual = traded_perpetual["price"]

                            traded_side_future = traded_future["side"]
                            traded_side_perpetual = traded_perpetual["side"]
                            traded_price_future = traded_future["price"]

                            exit_side = determine_exit_side_combo_auto(
                                traded_price_future,
                                traded_price_perpetual,
                                traded_side_future,
                                traded_side_perpetual,
                            )

                            if exit_side:
                                exit_params = {}
                                exit_params.update({"type": "limit"})
                                exit_params.update({"size": abs(traded_perpetual_size)})
                                exit_params.update({"entry_price": current_premium})
                                exit_params.update(
                                    {
                                        "label": f"{strategy_label}-closed-{label_integer}"
                                    }
                                )
                                exit_params.update(
                                    {"instrument_name": instrument_name_combo}
                                )
                                exit_params.update({"side": (exit_side)})

                                order_allowed = True

        return dict(
            order_allowed=order_allowed,
            order_parameters=([] if order_allowed == False else exit_params),
        )

    async def is_send_open_order_constructing_manual_combo_allowed(
        self,
        ticker_future: dict,
        instrument_attributes_futures: list,
        notional: float,
        monthly_target_profit: float,
        average_movement: float,
        basic_ticks_for_average_meovement: int,
        max_order_currency: int,
        market_condition: dict,
    ) -> dict:
        """ """

        order_allowed = False
        server_time = self.server_time
        my_trades = self.my_trades_currency_strategy
        my_trades_long = [o for o in my_trades if "sell" in o["side"]]
        my_trades_short = [o for o in my_trades if "buy" in o["side"]]

        my_trades_long_timestamp = (
            0 if my_trades_long == [] else max([o["timestamp"] for o in my_trades_long])
        )
        my_trades_short_timestamp = (
            0
            if my_trades_short == []
            else max([o["timestamp"] for o in my_trades_short])
        )

        delta = self.delta
        strategy_params: dict = self.strategy_parameters
        log.warning(f"constructing_manual_combo")
        # log.warning (f"{ticker_future}")

        instrument_name_future = ticker_future["instrument_name"]

        ask_price_future = ticker_future["best_ask_price"]

        orders_currency = self.orders_currency_strategy

        orders_instrument_future: list = [
            o for o in orders_currency if instrument_name_future in o["instrument_name"]
        ]

        log.info(f"orders_currency_strategy {orders_currency}")

        orders_instrument_future_open_all: list = [
            o
            for o in orders_currency
            if "PERPETUAL" not in o["instrument_name"]
            and "-FS-" not in o["instrument_name"]
            and "open" in o["label"]
        ]
        log.info(
            f"orders_instrument_future_open_all {orders_instrument_future_open_all}"
        )

        len_orders_instrument_future_open_all = (
            0
            if orders_instrument_future_open_all == []
            else len(orders_instrument_future_open_all)
        )

        orders_instrument_future_open: list = [
            o for o in orders_instrument_future if "open" in o["label"]
        ]

        orders_instrument_open: list = [
            o
            for o in orders_instrument_future_open
            if instrument_name_future in o["instrument_name"]
        ]

        len_orders_instrument: list = (
            0 if not orders_instrument_open else len(orders_instrument_open)
        )

        bullish, strong_bullish, weak_bullish = (
            market_condition["bullish"],
            market_condition["strong_bullish"],
            market_condition["weak_bullish"],
        )
        bearish, strong_bearish, weak_bearish = (
            market_condition["bearish"],
            market_condition["strong_bearish"],
            market_condition["weak_bearish"],
        )

        if delta == 0:

            bid_price_perpetual = self.ticker_perpetual["best_bid_price"]

            contango = is_contango(
                ask_price_future,
                bid_price_perpetual,
            )
            log.debug(
                f"contango {contango} len_orders_instrument_future_open_all {len_orders_instrument_future_open_all}"
            )

            if len_orders_instrument == 0 and contango:

                basic_size = determine_opening_size(
                    instrument_name_future,
                    instrument_attributes_futures,
                    notional,
                    monthly_target_profit,
                    average_movement,
                    basic_ticks_for_average_meovement,
                )

                label_open: str = get_label("open", self.strategy_label)

                # provide placeholder for params
                params = {}

                params.update({"instrument_name": instrument_name_future})
                params.update({"size": basic_size})
                params.update({"label": label_open})

                # default type: limit
                params.update({"type": "limit"})

                log.warning(
                    f"""PERPETUAL in instrument_name_future {"PERPETUAL" in instrument_name_future} 
                             PERPETUAL not in instrument_name_future {"PERPETUAL" not in instrument_name_future}"""
                )

                bullish_situation = strong_bullish or bullish or weak_bullish

                bearish_situation = strong_bearish or bearish or weak_bearish

                ONE_SECOND = 1000
                ONE_MINUTE = ONE_SECOND * 60

                last_buy_exceed_time_threshold = (
                    server_time - my_trades_long_timestamp
                    > strategy_params["waiting_minute_before_cancel"] * ONE_MINUTE
                )
                no_outstanding_long_position = len(my_trades_long) < 1

                last_sell_exceed_time_threshold = (
                    server_time - my_trades_short_timestamp
                    > strategy_params["waiting_minute_before_cancel"] * ONE_MINUTE
                )

                no_outstanding_short_position = len(my_trades_short) < 1

                log.error(
                    f"last_buy_exceed_time_threshold {last_buy_exceed_time_threshold} no_outstanding_long_position {no_outstanding_long_position}"
                )

                log.debug(
                    f"last_sell_exceed_time_threshold {last_sell_exceed_time_threshold} no_outstanding_short_position {no_outstanding_short_position}"
                )

                if (
                    bullish_situation
                    and "PERPETUAL" in instrument_name_future
                    and len_orders_instrument_future_open_all == 0
                    and (last_buy_exceed_time_threshold or no_outstanding_long_position)
                ):

                    order_allowed = True

                    params.update({"entry_price": bid_price_perpetual})

                    params.update({"side": "buy"})

                if (
                    bearish_situation
                    and "PERPETUAL" not in instrument_name_future
                    and len_orders_instrument_future_open_all < max_order_currency
                    and (
                        last_sell_exceed_time_threshold or no_outstanding_short_position
                    )
                ):

                    order_allowed = True

                    params.update({"entry_price": ask_price_future})

                    params.update({"side": "sell"})

        return dict(
            order_allowed=order_allowed,
            order_parameters=[] if order_allowed == False else params,
        )

    async def is_send_contra_order_for_unpaired_transaction_allowed(
        self,
        ticker_selected_transaction: dict,
        instrument_attributes_futures: list,
        take_profit_threshold_original: float,
        selected_transaction: dict,
        waiting_time_for_selected_transaction: bool,
        random_instruments_name: list,
        strategy_params: list = None,
    ) -> dict:
        """ """

        ticker_perpetual = self.ticker_perpetual

        log.warning(f"contra order")

        log.critical(selected_transaction)

        delta = self.delta

        strategy_label = self.strategy_label

        # provide placeholder for result
        result = dict(order_allowed=False, order_parameters=[])

        if selected_transaction:

            label_integer = get_label_integer(selected_transaction["label"])

            # order was not existed yet
            orders_currency = self.orders_currency_strategy

            log.info(f"orders_currency {orders_currency}")

            label_integer_in_orders_currency: list = [
                o for o in orders_currency if str(label_integer) in o["label"]
            ]

            log.info(
                f"label_integer_in_orders_currency {label_integer_in_orders_currency} not label_integer_in_orders_currency"
            )

            if not label_integer_in_orders_currency:

                instrument_side = selected_transaction["side"]

                instrument_name_transaction = selected_transaction["instrument_name"]

                tp_threshold = modified_tp_threshold(
                    instrument_attributes_futures,
                    take_profit_threshold_original,
                    instrument_name_transaction,
                )

                order_params_opening = get_basic_opening_parameters(strategy_label)

                if instrument_side == "buy":

                    result = await self.contra_order_for_unpaired_transaction_buy_side(
                        strategy_label,
                        tp_threshold,
                        delta,
                        order_params_opening,
                        orders_currency,
                        selected_transaction,
                        ticker_selected_transaction,
                        ticker_perpetual,
                        random_instruments_name,
                        waiting_time_for_selected_transaction,
                    )

                if instrument_side == "sell":

                    result = await self.contra_order_for_unpaired_transaction_sell_side(
                        strategy_label,
                        tp_threshold,
                        delta,
                        order_params_opening,
                        orders_currency,
                        selected_transaction,
                        ticker_selected_transaction,
                        ticker_perpetual,
                        waiting_time_for_selected_transaction,
                    )

        log.error(f"result {result} ")
        return result

    async def contra_order_for_unpaired_transaction_sell_side(
        self,
        strategy_label,
        tp_threshold: dict,
        delta: float,
        params,
        orders_currency,
        selected_transaction: dict,
        ticker_selected_transaction: bool,
        ticker_perpetual,
        waiting_time_for_selected_transaction,
    ) -> dict:
        """ """

        log.info("transaction sell side")

        counter_side = "buy"

        order_allowed = False

        reduce_only = self.strategy_parameters["reduce_only"]  # [0]

        selected_transaction_price = selected_transaction["price"]

        selected_transaction_size = abs(selected_transaction["amount"])

        basic_size = selected_transaction["amount"]

        contra_order_will_reduce_delta = is_contra_order_will_reduce_delta(
            delta, selected_transaction_size, counter_side
        )

        if contra_order_will_reduce_delta:

            instrument_name_perpetual: str = ticker_perpetual["instrument_name"]

            instrument_name_transaction: str = selected_transaction["instrument_name"]

            selected_transaction_side: str = selected_transaction["side"]

            orders_instrument_perpetual: list = [
                o
                for o in orders_currency
                if instrument_name_perpetual in o["instrument_name"]
            ]

            bid_price_perpetual, ask_price_perpetual = (
                ticker_perpetual["best_bid_price"],
                ticker_perpetual["best_ask_price"],
            )

            orders_instrument_transaction: list = [
                o
                for o in orders_currency
                if instrument_name_transaction in o["instrument_name"]
            ]

            orders_instrument_transaction_closed: list = [
                o for o in orders_instrument_transaction if "closed" in o["label"]
            ]

            orders_instrument_transaction_net: int = (
                0
                if orders_instrument_transaction == []
                else sum([o["amount"] for o in orders_instrument_transaction])
            )

            orders_instrument_perpetual_open: list = [
                o for o in orders_instrument_perpetual if "open" in o["label"]
            ]

            len_orders_instrument_transaction: int = (
                0
                if not orders_instrument_transaction
                else len(orders_instrument_transaction)
            )

            len_orders_instrument_perpetual: int = (
                0
                if not orders_instrument_perpetual
                else len(orders_instrument_perpetual_open)
            )

            params.update({"side": counter_side})

            label_integer = get_label_integer(selected_transaction["label"])

            instrument_current_size = sum(
                [
                    o["amount"]
                    for o in (self.my_trades_currency_strategy)
                    if instrument_name_transaction in o["instrument_name"]
                ]
            )

            log.error(
                f"instrument_current_size {instrument_current_size} orders_instrument_transaction_net {orders_instrument_transaction_net} selected_transaction_size {selected_transaction_size}"
            )

            sum_order_under_closed_label = basic.sum_order_under_closed_label_int(
                orders_instrument_transaction_closed, label_integer
            )

            net_size = basic_size + sum_order_under_closed_label

            size_abs = abs(basic_size)

            params.update({"size": size_abs})

            size = size_abs * basic.ensure_sign_consistency(counter_side)

            closing_size_ok = (
                basic.check_if_next_closing_size_will_not_exceed_the_original(
                    basic_size, net_size, size
                )
            )

            # log.error (f"basic_size {basic_size} net_size {net_size} size {size} sum_order_under_closed_label {sum_order_under_closed_label}")

            if (
                "PERPETUAL" not in instrument_name_transaction
                and ticker_selected_transaction
                and closing_size_ok
            ):

                bid_price_selected_transaction = ticker_selected_transaction[
                    "best_bid_price"
                ]
                ask_price_selected_transaction = ticker_selected_transaction[
                    "best_ask_price"
                ]

                transaction_in_profit = basic.profit_usd_has_exceed_target(
                    tp_threshold,
                    selected_transaction_price,
                    ask_price_selected_transaction,
                    basic_size,
                    selected_transaction_side,
                )

                target_price = selected_transaction_price - (
                    selected_transaction_price * tp_threshold
                )

                transaction_in_profit = bid_price_selected_transaction < target_price

                log.error(
                    f"transaction_in_profit {transaction_in_profit} bid_price_selected_transaction {bid_price_selected_transaction} {selected_transaction_price} {(selected_transaction_price - selected_transaction_price * tp_threshold)}"
                )

                if transaction_in_profit:

                    if len_orders_instrument_transaction == 0:

                        order_allowed = True

                        params.update({"instrument_name": instrument_name_transaction})

                        label = f"{strategy_label}-closed-{label_integer}"

                        params.update({"label": label})
                        params.update({"entry_price": bid_price_selected_transaction})

                # using perpetual to balancing delta
                else:

                    sum_orders_instrument_perpetual_open = (
                        0
                        if orders_instrument_perpetual_open == []
                        else sum(
                            [o["amount"] for o in orders_instrument_perpetual_open]
                        )
                    )

                    log.debug(
                        f"sum_orders_instrument_perpetual_open {sum_orders_instrument_perpetual_open} "
                    )
                    log.error(
                        f"selected_transaction_price {selected_transaction_price} <= bid_price_perpetual {bid_price_perpetual}  {selected_transaction_price <= bid_price_perpetual} "
                    )
                    log.warning(
                        f"waiting_time_for_selected_transaction {waiting_time_for_selected_transaction} selected_transaction_price > bid_price_perpetual {selected_transaction_price > bid_price_perpetual}"
                    )

                    if (
                        len_orders_instrument_perpetual == 0
                        and sum_orders_instrument_perpetual_open < abs(delta)
                        and delta <= 0
                    ):

                        if waiting_time_for_selected_transaction:
                            order_allowed = True

                            # pairing with perpetual
                            if selected_transaction_price > bid_price_perpetual:
                                params.update({"label": selected_transaction["label"]})

                            params.update(
                                {"instrument_name": instrument_name_perpetual}
                            )

                            params.update({"entry_price": bid_price_perpetual})

                            orders_instrument: list = [
                                o
                                for o in orders_instrument_perpetual_open
                                if instrument_name_perpetual in o["instrument_name"]
                            ]

            if "PERPETUAL" in instrument_name_transaction and closing_size_ok:

                transaction_in_profit = basic.profit_usd_has_exceed_target(
                    tp_threshold,
                    selected_transaction_price,
                    ask_price_perpetual,
                    basic_size,
                    selected_transaction_side,
                )

                log.error(
                    f"transaction_in_profit {transaction_in_profit} bid_price_perpetual {bid_price_perpetual} {selected_transaction_price} {(selected_transaction_price - selected_transaction_price * tp_threshold)}"
                )

                # sell immediately when in profit
                if transaction_in_profit:

                    if len_orders_instrument_perpetual == 0:

                        order_allowed = True

                        params.update({"instrument_name": instrument_name_perpetual})

                        label = f"{strategy_label}-closed-{label_integer}"
                        log.error(f"label {label}")

                        params.update({"label": label})
                        params.update({"entry_price": bid_price_perpetual})

        return dict(
            order_allowed=order_allowed,
            order_parameters=[] if order_allowed == False else params,
        )

    async def contra_order_for_unpaired_transaction_buy_side(
        self,
        strategy_label,
        tp_threshold: dict,
        delta: float,
        params,
        orders_currency,
        selected_transaction: dict,
        ticker_selected_transaction: dict,
        ticker_perpetual: dict,
        random_instruments_name: list,
        waiting_time_for_selected_transaction,
    ) -> dict:
        """ """

        log.info("transaction buy side")

        counter_side = "sell"

        order_allowed = False

        reduce_only = self.strategy_parameters["reduce_only"]  # [0]

        selected_transaction_price = selected_transaction["price"]

        selected_transaction_size = abs(selected_transaction["amount"])

        basic_size = selected_transaction["amount"]

        contra_order_will_reduce_delta = is_contra_order_will_reduce_delta(
            delta,
            selected_transaction_size,
            counter_side,
        )

        if contra_order_will_reduce_delta:

            instrument_name_perpetual = ticker_perpetual["instrument_name"]

            instrument_name_transaction = selected_transaction["instrument_name"]

            selected_transaction_side = selected_transaction["side"]

            bid_price_selected_transaction = ticker_selected_transaction[
                "best_bid_price"
            ]
            ask_price_selected_transaction = ticker_selected_transaction[
                "best_ask_price"
            ]
            # if delta < 0:
            size_abs = abs(basic_size)

            size = size_abs * basic.ensure_sign_consistency(counter_side)

            params.update({"side": counter_side})

            params.update({"size": size_abs})

            orders_instrument_transaction: list = [
                o
                for o in orders_currency
                if instrument_name_transaction in o["instrument_name"]
            ]

            orders_instrument_perpetual: list = [
                o
                for o in orders_currency
                if instrument_name_perpetual in o["instrument_name"]
            ]

            orders_instrument_transaction_closed: list = [
                o for o in orders_instrument_transaction if "closed" in o["label"]
            ]

            orders_instrument_perpetual_closed: list = [
                o for o in orders_instrument_perpetual if "closed" in o["label"]
            ]

            log.info(f"contra_order_will_reduce_delta {contra_order_will_reduce_delta}")

            log.info(
                f"delta {delta} selected_transaction_size {selected_transaction_size} counter_side {counter_side}"
            )

            selected_transaction_label = selected_transaction["label"]

            label_integer = get_label_integer(selected_transaction_label)

            # ask_price_future = ticker_future ["best_ask_price"]
            bid_price_perpetual, ask_price_perpetual = (
                ticker_perpetual["best_bid_price"],
                ticker_perpetual["best_ask_price"],
            )

            instrument_name_future = random_instruments_name[0]

            ticker_instrument = reading_from_pkl_data("ticker", instrument_name_future)

            if "PERPETUAL" in instrument_name_transaction:

                if delta > 0:
                    transaction_in_profit = bid_price_perpetual >= (
                        selected_transaction_price
                    )

                else:
                    transaction_in_profit = basic.profit_usd_has_exceed_target(
                        tp_threshold,
                        selected_transaction_price,
                        bid_price_perpetual,
                        basic_size,
                        selected_transaction_side,
                    )

                log.error(
                    f"transaction_in_profit {transaction_in_profit} orders_instrument_transaction_closed == 0 {orders_instrument_transaction_closed == 0}"
                )
                log.debug(
                    f"orders_instrument_transaction {orders_instrument_transaction}"
                )

                if transaction_in_profit:

                    sum_order_under_closed_label = (
                        basic.sum_order_under_closed_label_int(
                            orders_instrument_perpetual_closed, label_integer
                        )
                    )

                    net_size = basic_size + sum_order_under_closed_label

                    closing_size_ok = (
                        basic.check_if_next_closing_size_will_not_exceed_the_original(
                            basic_size, net_size, size
                        )
                    )

                    if not orders_instrument_transaction_closed and closing_size_ok:

                        params.update({"instrument_name": instrument_name_perpetual})

                        label_integer = get_label_integer(selected_transaction["label"])

                        closed_label = f"{strategy_label}-closed-{label_integer}"
                        log.error(f"closed_label {closed_label}")

                        params.update({"label": closed_label})
                        params.update({"entry_price": ask_price_perpetual})

                        order_allowed = True

                else:

                    log.debug(f"ticker_instrument {ticker_instrument}")
                    if ticker_instrument and "PERPETUAL" not in instrument_name_future:

                        ticker_instrument = ticker_instrument[0]

                        orders_instrument_future: list = [
                            o
                            for o in orders_currency
                            if instrument_name_future in o["instrument_name"]
                        ]

                        len_orders_instrument_future: int = (
                            0
                            if not orders_instrument_future
                            else len(orders_instrument_future)
                        )

                        log.debug(
                            f"len_orders_instrument_future {len_orders_instrument_future}"
                        )
                        if len_orders_instrument_future == 0:

                            if waiting_time_for_selected_transaction:
                                order_allowed = True

                                ask_price = ticker_instrument["best_ask_price"]

                                # creating paired combo, else independent transaction
                                if ask_price > selected_transaction_price:
                                    params.update({"label": selected_transaction_label})

                                params.update({"entry_price": ask_price})

                                params.update(
                                    {"instrument_name": instrument_name_future}
                                )

                                order_allowed = True

            if (
                "PERPETUAL" not in instrument_name_transaction
                and ticker_selected_transaction
            ):

                sum_order_under_closed_label = basic.sum_order_under_closed_label_int(
                    orders_instrument_transaction_closed, label_integer
                )

                net_size = basic_size + sum_order_under_closed_label

                closing_size_ok = (
                    basic.check_if_next_closing_size_will_not_exceed_the_original(
                        basic_size, net_size, size
                    )
                )

                log.warning(
                    f"closing_size_ok {closing_size_ok} { sum_order_under_closed_label}"
                )
                log.info(
                    f"orders_instrument_transaction_closed  { orders_instrument_transaction_closed}"
                )

                if closing_size_ok:

                    if delta > 0:
                        transaction_in_profit = bid_price_selected_transaction >= (
                            selected_transaction_price
                        )

                    else:
                        transaction_in_profit = basic.profit_usd_has_exceed_target(
                            tp_threshold,
                            selected_transaction_price,
                            bid_price_selected_transaction,
                            basic_size,
                            selected_transaction_side,
                        )

                    log.error(
                        f"transaction_in_profit {transaction_in_profit} orders_instrument_transaction_closed == 0 {orders_instrument_transaction_closed == 0}"
                    )
                    log.info(
                        f"instrument_name_transaction {instrument_name_transaction}"
                    )
                    log.debug(
                        f"orders_instrument_transaction {orders_instrument_transaction}"
                    )
                    # log.warning (f"orders_currency {orders_currency}")

                    # using the same instrument as transaction instrument
                    if transaction_in_profit:

                        if not orders_instrument_transaction_closed:

                            params.update(
                                {"instrument_name": instrument_name_transaction}
                            )

                            closed_label = f"{strategy_label}-closed-{label_integer}"

                            params.update({"label": closed_label})
                            params.update(
                                {"entry_price": ask_price_selected_transaction}
                            )

                            order_allowed = True

        return dict(
            order_allowed=order_allowed,
            order_parameters=[] if order_allowed == False else params,
        )



# # -*- coding: utf-8 -*-

# built ins
import asyncio

# installed
from loguru import logger as log

# user defined formula
from streaming_helper.db_management import sqlite_management as db_mgt
from streaming_helper.strategies.deribit.cash_carry import combo_auto
from streaming_helper.utilities import string_modification as str_mod


def waiting_time_has_expired(
    strategy_params: dict,
    future_trade: dict,
    perpetual_trade: dict,
    server_time: int,
) -> bool:
    """ """

    ONE_SECOND = 1000

    ONE_MINUTE = ONE_SECOND * 60

    waiting_minute_before_cancel = (
        strategy_params["waiting_minute_before_relabelling"] * ONE_MINUTE
    )

    # log.debug (f"waiting_minute_before_cancel {waiting_minute_before_cancel}")

    timestamp_perpetual: int = perpetual_trade["timestamp"]

    waiting_time_for_perpetual_order: bool = (
        combo_auto.check_if_minimum_waiting_time_has_passed(
            waiting_minute_before_cancel,
            timestamp_perpetual,
            server_time,
        )
    )

    timestamp_future: int = future_trade["timestamp"]

    waiting_time_for_future_order: bool = (
        combo_auto.check_if_minimum_waiting_time_has_passed(
            waiting_minute_before_cancel,
            timestamp_future,
            server_time,
        )
    )

    return waiting_time_for_perpetual_order and waiting_time_for_future_order


def my_trades_currency_strategy_with_no_blanks(
    my_trades_currency: list,
    strategy: str,
) -> list:
    """ """

    my_trades_currency_active_with_no_blanks = (
        []
        if my_trades_currency == []
        else [o for o in my_trades_currency if o["label"] is not None]
    )

    my_trades_currency_strategy = [
        o
        for o in my_trades_currency_active_with_no_blanks
        if strategy in o["label"]
        and "closed" not in o["label"]
        and "Auto" not in o["label"]
    ]
    return my_trades_currency_strategy


def get_redundant_ids(
    my_trades_currency: list,
    strategy: str,
) -> list:
    """ """

    my_trades_currency_strategy = my_trades_currency_strategy_with_no_blanks(
        my_trades_currency, strategy
    )

    if my_trades_currency_strategy:

        instrument_names = str_mod.remove_redundant_elements(
            [o["instrument_name"] for o in my_trades_currency_strategy]
        )

        if instrument_names:

            for instrument_name in instrument_names:

                my_trade_instrument_name = [
                    o
                    for o in my_trades_currency_strategy
                    if instrument_name in o["instrument_name"]
                ]

                if my_trade_instrument_name:

                    my_trades_label = str_mod.remove_redundant_elements(
                        [(o["label"]) for o in my_trade_instrument_name]
                    )

                    result = []
                    for label in my_trades_label:
                        len_label = len(
                            [
                                o["label"]
                                for o in my_trades_currency_strategy
                                if label in o["label"]
                            ]
                        )
                        if len_label > 1:
                            result.append(label)

                    return result


def get_single_transaction(
    my_trades_currency: list,
    strategy: str,
) -> list:
    """ """

    my_trades_currency_strategy = my_trades_currency_strategy_with_no_blanks(
        my_trades_currency, strategy
    )

    if my_trades_currency_strategy:

        my_trades_label = str_mod.remove_redundant_elements(
            [(o["label"]) for o in my_trades_currency_strategy]
        )

        result = []
        for label in my_trades_label:

            label_integer = str_mod.parsing_label(label)["int"]

            transaction_under_label_integer = [
                o for o in my_trades_currency_strategy if label_integer in o["label"]
            ]

            # additional filter
            sum_transaction_under_label_integer = sum(
                [o["amount"] for o in transaction_under_label_integer]
            )

            transaction_under_label_integer_len = len(transaction_under_label_integer)
            # log.info (f" {transaction_under_label_integer} sum_transaction_under_label_integer {sum_transaction_under_label_integer} transaction_under_label_integer_len {transaction_under_label_integer_len}")
            if transaction_under_label_integer_len == 1:

                result.append(transaction_under_label_integer[0])

        return result


async def updating_db_with_new_label(
    trade_db_table: str,
    archive_db_table: str,
    trade_id: str,
    filter: str,
    new_label: str,
) -> None:
    """ """

    await db_mgt.update_status_data(
        archive_db_table, "label", filter, trade_id, new_label, "="
    )

    await db_mgt.update_status_data(
        trade_db_table, "label", filter, trade_id, new_label, "="
    )


async def relabelling_double_ids(
    trade_db_table: str,
    archive_db_table: str,
    my_trades_currency_active: list,
) -> None:
    """ """

    from strategies.basic_strategy import get_label

    strategy = "futureSpread"

    relabelling = False

    my_trades_currency_strategy = my_trades_currency_strategy_with_no_blanks(
        my_trades_currency_active, strategy
    )

    if my_trades_currency_strategy:

        instrument_names = str_mod.remove_redundant_elements(
            [o["instrument_name"] for o in my_trades_currency_strategy]
        )

        if instrument_names:

            for instrument_name in instrument_names:

                my_trade_instrument_name = [
                    o
                    for o in my_trades_currency_strategy
                    if instrument_name in o["instrument_name"]
                ]

                redundant_ids = get_redundant_ids(
                    my_trade_instrument_name,
                    strategy,
                )

                if redundant_ids:

                    log.error(f"redundant_ids {redundant_ids}")

                    for label in redundant_ids:
                        log.error(f"label {label}")

                        trade_ids = [
                            o["trade_id"]
                            for o in my_trade_instrument_name
                            if label in o["label"]
                        ]

                        if trade_ids:

                            for trade_id in trade_ids:

                                log.warning(f"trade_id {trade_id}")

                                filter = "trade_id"

                                new_label: str = get_label("open", strategy)

                                await updating_db_with_new_label(
                                    trade_db_table,
                                    archive_db_table,
                                    trade_id,
                                    filter,
                                    new_label,
                                )

                                relabelling = True

                                break
    return relabelling


async def pairing_single_label(
    strategy_attributes: list,
    archive_db_table: str,
    my_trades_currency_active: dict,
    server_time: int,
) -> None:
    """ """

    paired_success = False

    strategy = "futureSpread"

    single_label_transaction = get_single_transaction(
        my_trades_currency_active, strategy
    )

    # log.warning (f"single_label_transaction {single_label_transaction}")

    if single_label_transaction:

        my_trades_amount = str_mod.remove_redundant_elements(
            [abs(o["amount"]) for o in single_label_transaction]
        )

        strategy_params = strategy_params = [
            o for o in strategy_attributes if o["strategy_label"] == strategy
        ][0]

        # log.error(f"my_trades_amount {my_trades_amount}")

        for amount in my_trades_amount:

            # log.error(f"amount {amount}")

            my_trades_with_the_same_amount = [
                o for o in single_label_transaction if amount == abs(o["amount"])
            ]

            my_trades_with_the_same_amount_label_perpetual = [
                o
                for o in my_trades_with_the_same_amount
                if "PERPETUAL" in o["instrument_name"]
            ]

            my_trades_with_the_same_amount_label_non_perpetual = [
                o
                for o in my_trades_with_the_same_amount
                if "PERPETUAL" not in o["instrument_name"]
            ]

            my_trades_with_the_same_amount_label_non_perpetual_instrument_name = (
                str_mod.remove_redundant_elements(
                    [
                        o["instrument_name"]
                        for o in my_trades_with_the_same_amount_label_non_perpetual
                    ]
                )
            )

            # log.error(
            #    f"my_trades_with_the_same_amount_label_non_perpetual_instrument_name {my_trades_with_the_same_amount_label_non_perpetual_instrument_name}"
            # )

            for (
                instrument_name_future
            ) in my_trades_with_the_same_amount_label_non_perpetual_instrument_name:

                my_trades_with_the_same_amount_label_future = [
                    o
                    for o in my_trades_with_the_same_amount_label_non_perpetual
                    if instrument_name_future in o["instrument_name"]
                ]

                my_trades_future_sorted = str_mod.sorting_list(
                    my_trades_with_the_same_amount_label_future, "price", True
                )

                if my_trades_future_sorted:

                    future_trade = my_trades_future_sorted[0]
                    price_future = future_trade["price"]

                    my_trades_perpetual_with_lower_price = [
                        o
                        for o in my_trades_with_the_same_amount_label_perpetual
                        if o["price"] < price_future and o["amount"] > 0
                    ]

                    my_trades_perpetual_with_lower_price_sorted = str_mod.sorting_list(
                        my_trades_perpetual_with_lower_price, "price", False
                    )

                    # log.error(f"my_trades_future_sorted {my_trades_future_sorted}")

                    if my_trades_perpetual_with_lower_price_sorted:

                        # log.debug (f"my_trades_perpetual_with_lower_price_sorted {my_trades_perpetual_with_lower_price_sorted}")

                        perpetual_trade = my_trades_perpetual_with_lower_price_sorted[0]

                        log.warning(future_trade)
                        log.debug(perpetual_trade)

                        waiting_time_expired = waiting_time_has_expired(
                            strategy_params,
                            future_trade,
                            perpetual_trade,
                            server_time,
                        )

                        if waiting_time_expired:

                            side_perpetual = perpetual_trade["side"]
                            side_future = future_trade["side"]

                            filter = "trade_id"
                            trade_id = perpetual_trade[filter]
                            new_label = future_trade["label"]

                            # market contango
                            if my_trades_perpetual_with_lower_price_sorted:

                                # market contango
                                if side_future == "sell" and side_perpetual == "buy":

                                    await db_mgt.update_status_data(
                                        archive_db_table,
                                        "label",
                                        filter,
                                        trade_id,
                                        new_label,
                                        "=",
                                    )

                                    log.warning(future_trade)
                                    log.debug(perpetual_trade)
                                    log.debug(new_label)

                                    paired_success = True

                                    break

    return paired_success



# -*- coding: utf-8 -*-

# built ins
import asyncio

# installed
from dataclassy import dataclass

# user defined formula
from streaming_helper.utilities import time_modification as time_mod
from streaming_helper.restful_api import connector


def get_basic_https() -> str:
    return f"https://www.deribit.com/api/v2/"


def get_currencies_end_point() -> str:
    return f"public/get_currencies?"

async def get_currencies() -> str:
    return await connector.get_connected(
            get_basic_https(),
            get_currencies_end_point(),
        )

def get_server_time_end_point() -> str:
    return f"public/get_time?"


def get_instruments_end_point(currency) -> str:
    return f"public/get_instruments?currency={currency.upper()}"

async def get_instruments(currency) -> str:
    return await connector.get_connected(
            get_basic_https(),
            get_instruments_end_point(currency),
        )


def get_tickers_end_point(instrument_name: str) -> str:

    return f"public/ticker?instrument_name={instrument_name}"


def get_tradingview_chart_data_end_point() -> str:
    return f"get_tradingview_chart_data?"


def get_ohlc_end_point(
    endpoint_tradingview: str,
    instrument_name: str,
    resolution: int,
    qty_or_start_time_stamp: int,
    provided_end_timestamp: int = None,
    qty_as_start_time_stamp: bool = False,
) -> str:

    now_unix = time_mod.get_now_unix_time()

    # start timestamp is provided
    start_timestamp = qty_or_start_time_stamp

    # recalculate start timestamp using qty as basis point
    if qty_as_start_time_stamp:
        start_timestamp = now_unix - (60000 * resolution) * qty_as_start_time_stamp

    if provided_end_timestamp:
        end_timestamp = provided_end_timestamp
    else:
        end_timestamp = now_unix

    return f"{endpoint_tradingview}end_timestamp={end_timestamp}&instrument_name={instrument_name}&resolution={resolution}&start_timestamp={start_timestamp}"


def get_json_payload(
    endpoint: str,
    params: dict,
) -> dict:

    id = id_numbering(
        endpoint,
        endpoint,
    )

    return {
        "jsonrpc": "2.0",
        "id": id,
        "method": f"{endpoint}",
        "params": params,
    }


def get_open_orders_end_point() -> str:
    return f"private/get_open_orders"


def get_open_orders_params(
    kind: str,
    type: str,
) -> dict:
    return {
        "kind": kind,
        "type": type,
    }


def get_subaccounts_end_point() -> str:
    return f"private/get_subaccounts"


def get_subaccounts_params(
    with_portfolio: bool = True,
) -> dict:
    return {
        "with_portfolio": with_portfolio,
    }


def get_subaccounts_details_end_point() -> str:
    return f"private/get_subaccounts_details"


def get_subaccounts_details_params(
    currency: str,
    with_open_orders: bool = True,
) -> dict:
    return {
        "currency": currency,
        "with_open_orders": with_open_orders,
    }

def get_end_point_based_on_side(side: str) -> str:

    if side == "buy":
        return "private/buy"

    if side == "sell":
        return "private/sell"



@dataclass(unsafe_hash=True, slots=True)
class SendApiRequest:
    """ """

    client_id: str
    client_secret: str
    
    
    async def get_open_orders(
        self,
        kind: str,
        type: str,  
        ):

        result = await connector.get_connected(
            get_basic_https(),
            get_open_orders_end_point(),
            self.client_id,
            self.client_secret,
            get_open_orders_params(
                kind, 
                type,
                )
        ) 

        return result

    
    async def get_subaccounts(
    self,
    with_portfolio: bool = True
    ) -> list:
    
        sub_account = await connector.get_connected(
        get_basic_https(),
        get_subaccounts_end_point(),
        self.client_id,
        self.client_secret,
        get_subaccounts_params(with_portfolio))

        return sub_account["result"]
    
    
    async def get_subaccounts_details(
    self,
    currency: str,
    with_open_orders: bool = True
    ) -> list:
    
        """

        currency= BTC/ETH
        example= https://www.deribit.com/api/v2/private/get_subaccounts_details?currency=BTC&with_open_orders=true


        result_sub_account["result"]

        Returns example:

         [
             {
                'positions': [
                     {
                         'estimated_liquidation_price': None,
                         'size_currency': -0.031537551,
                         'total_profit_loss': -0.005871738,
                         'realized_profit_loss': 0.0,
                         'floating_profit_loss': -0.002906191,
                         'leverage': 25,
                         'average_price': 74847.72,
                         'delta': -0.031537551,
                         'mark_price': 88783.05,
                         'settlement_price': 81291.98,
                         'instrument_name': 'BTC-15NOV24',
                         'index_price': 88627.96,
                         'direction': 'sell',
                         'open_orders_margin': 0.0,
                         'initial_margin': 0.001261552,
                         'maintenance_margin': 0.000630801,
                         'kind': 'future',
                         'size': -2800.0
                     },
                     {
                         'estimated_liquidation_price': None,
                         'size_currency': -0.006702271,
                         'total_profit_loss': -0.001912148,
                         'realized_profit_loss': 0.0,
                         'floating_profit_loss': -0.000624473,
                         'leverage': 25,
                         'average_price': 69650.67,
                         'delta': -0.006702271,
                         'mark_price': 89521.9,
                         'settlement_price': 81891.77,
                         'instrument_name': 'BTC-29NOV24',
                         'index_price': 88627.96,
                         'direction': 'sell',
                         'open_orders_margin': 0.0,
                         'initial_margin': 0.000268093,
                         'maintenance_margin': 0.000134048,
                         'kind': 'future',
                         'size': -600.0
                     },
                     {
                         'estimated_liquidation_price': None,
                         'size_currency': 0.036869785,
                         'realized_funding': -2.372e-05,
                         'total_profit_loss': 0.005782196,
                         'realized_profit_loss': 0.000591453,
                         'floating_profit_loss': 0.002789786,
                         'leverage': 50,
                         'average_price': 76667.01,
                         'delta': 0.036869785,
                         'interest_value': 0.2079087278497569,
                         'mark_price': 88690.51,
                         'settlement_price': 81217.47,
                         'instrument_name': 'BTC-PERPETUAL',
                         'index_price': 88627.96,
                         'direction': 'buy',
                         'open_orders_margin': 3.489e-06,
                         'initial_margin': 0.000737464,
                         'maintenance_margin': 0.000368766,
                         'kind': 'future',
                         'size': 3270.0
                     }
                     ],

                'open_orders': [
                     {
                         'is_liquidation': False,
                         'risk_reducing': False,
                         'order_type': 'limit',
                         'creation_timestamp': 1731390729846,
                         'order_state': 'open',
                         'reject_post_only': False,
                         'contracts': 1.0,
                         'average_price': 0.0,
                         'reduce_only': False,
                         'post_only': True,
                         'last_update_timestamp': 1731390729846,
                         'filled_amount': 0.0,
                         'replaced': False,
                         'mmp': False,
                         'web': False,
                         'api': True,
                         'instrument_name': 'BTC-PERPETUAL',
                         'amount': 10.0,
                         'order_id': '80616245864',
                         'max_show': 10.0,
                         'time_in_force': 'good_til_cancelled',
                         'direction': 'buy',
                         'price': 88569.5,
                         'label': 'hedgingSpot-closed-1731387973670'
                     }
                     ],

                'uid': 148510
                }
        ]

        """

        sub_account = await connector.get_connected(
        get_basic_https(),
        get_subaccounts_details_end_point(),
        self.client_id,
        self.client_secret,
        get_subaccounts_details_params(
        currency,
        with_open_orders,
        ))

        return sub_account["result"]
    
    async def get_transaction_log(
        self,
        currency: str,
        start_timestamp: int,
        count: int = 1000,
        query: str = "trade",
    ) -> list:
        """
        query:
            trade, maker, taker, open, close, liquidation, buy, sell,
            withdrawal, delivery, settlement, deposit, transfer,
            option, future, correction, block_trade, swap

        """


        result_transaction_log_to_result = await connector.get_connected(
        get_basic_https(),
        get_transaction_log_end_point(),
        self.client_id,
        self.client_secret,
        get_transaction_log_params(
        currency,
        start_timestamp,
        count,
        query,
        ))


        try:
            result = result_transaction_log_to_result["result"]

            return [] if not result else result["logs"]

        except:

            error = result_transaction_log_to_result["error"]
            
            message = error["message"]
            
            return message

    async def get_cancel_order_all(
        self,
        detailed: bool = False
        ):

        # Set endpoint
        endpoint: str = "private/cancel_all"

        params = {"detailed": False}


        result = await connector.get_connected(
        get_basic_https(),
        cancel_all_orders_end_point(),
        self.client_id,
        self.client_secret,
        cancel_all_orders_params(detailed))

        return result


    async def get_cancel_order_byOrderId(
        self,
        order_id: str,
    ) -> None:
        # Set endpoint

        result = await connector.get_connected(
        get_basic_https(),
        cancel_order_end_point(),
        self.client_id,
        self.client_secret,
        get_cancel_order_params(order_id))

        return result


    async def send_order(
        self,
        side: str,
        instrument,
        amount,
        label: str = None,
        price: float = None,
        type: str = "limit",
        otoco_config: list = None,
        linked_order_type: str = None,
        trigger_price: float = None,
        trigger: str = "last_price",
        time_in_force: str = "fill_or_kill",
        reduce_only: bool = False,
        post_only: bool = True,
        reject_post_only: bool = False,
    ) -> None:

        params = {}

        params.update({"instrument_name": instrument})
        params.update({"amount": amount})
        params.update({"label": label})
        params.update({"instrument_name": instrument})
        params.update({"type": type})

        if trigger_price is not None:

            params.update({"trigger": trigger})
            params.update({"trigger_price": trigger_price})
            params.update({"reduce_only": reduce_only})

        if "market" not in type:
            params.update({"price": price})
            params.update({"post_only": post_only})
            params.update({"reject_post_only": reject_post_only})

        if otoco_config:
            params.update({"otoco_config": otoco_config})

            if linked_order_type is not None:
                params.update({"linked_order_type": linked_order_type})
            else:
                params.update({"linked_order_type": "one_triggers_other"})

            params.update({"trigger_fill_condition": "incremental"})

            log.debug(f"params otoco_config {params}")

        result = None

        if side is not None:

            endpoint: str = get_end_point_based_on_side(side)

            result = await private_connection(
                self.sub_account_id,
                endpoint=endpoint,
                params=params,
            )

        return result

    async def get_open_orders(
        self,
        kind: str,
        type: str,
    ) -> list:

        # Set endpoint
        endpoint: str = "private/get_open_orders"

        params = {"kind": kind, "type": type}

        result_open_order = await private_connection(
            self.sub_account_id,
            endpoint=endpoint,
            params=params,
        )

        return result_open_order["result"]

    async def send_limit_order(
        self,
        params: dict,
    ) -> None:
        """ """

        # basic params
        log.info(f"params {params}")
        side = params["side"]
        instrument = params["instrument_name"]
        label_numbered = params["label"]
        size = params["size"]
        type = params["type"]
        limit_prc = params["entry_price"]

        try:
            otoco_config = params["otoco_config"]
        except:
            otoco_config = None

        try:
            linked_order_type = params["linked_order_type"]

        except:
            linked_order_type = None

        order_result = None

        if side != None:

            if type == "limit":  # limit has various state

                order_result = await self.send_order(
                    side,
                    instrument,
                    size,
                    label_numbered,
                    limit_prc,
                    type,
                    otoco_config,
                )

            else:

                trigger_price = params["trigger_price"]
                trigger = params["trigger"]

                order_result = await self.send_order(
                    side,
                    instrument,
                    size,
                    label_numbered,
                    limit_prc,
                    type,
                    otoco_config,
                    linked_order_type,
                    trigger_price,
                    trigger,
                )

        # log.warning(f"order_result {order_result}")

        if order_result != None and (
            "error" in order_result or "message" in order_result
        ):

            error = order_result["error"]
            message = error["message"]

            try:
                data = error["data"]
            except:
                data = message

            await tlgrm.telegram_bot_sendtext(
                f"message: {message}, \
                                         data: {data}, \
                                         (params: {params}"
            )

        log.warning(f"order_result {order_result}")
        return order_result
    
def get_user_trades_by_currency_end_point() -> str:
    return f"private/get_user_trades_by_currency"


def get_user_trades_by_currency_params(
    currency: str,
    kind: str = "any",
    count: int = 1000,
) -> dict:
    return {
        "currency": currency,
        "kind": kind,
        "count": count,
    }


def get_user_trades_by_instrument_and_time_end_point() -> str:
    return f"private/get_user_trades_by_instrument_and_time"


def get_user_trades_by_instrument_and_time_params(
    instrument_name: str,
    start_timestamp: int,
    count: int = 1000,
) -> dict:

    now_unix = time_mod.get_now_unix_time()

    return {
        "count": count,
        "end_timestamp": now_unix,
        "instrument_name": instrument_name,
        "start_timestamp": start_timestamp,
    }


def get_transaction_log_end_point() -> str:
    return f"private/get_transaction_log"


def get_transaction_log_params(
    currency: str,
    start_timestamp: int,
    count: int = 1000,
    query: str = "trade",
) -> dict:

    now_unix = time_mod.get_now_unix_time()

    return {
        "count": count,
        "currency": currency,
        "end_timestamp": now_unix,
        "query": query,
        "start_timestamp": start_timestamp,
    }


def send_orders_end_point(params: dict) -> str:

    return f"private/{params["side"]}"


def send_orders_params(
    side: str,
    instrument,
    amount,
    label: str = None,
    price: float = None,
    type: str = "limit",
    otoco_config: list = None,
    linked_order_type: str = None,
    trigger_price: float = None,
    trigger: str = "last_price",
    time_in_force: str = "fill_or_kill",
    reduce_only: bool = False,
    post_only: bool = True,
    reject_post_only: bool = False,
) -> str:

    params = {}

    params.update({"instrument_name": instrument})
    params.update({"amount": amount})
    params.update({"label": label})
    params.update({"instrument_name": instrument})
    params.update({"type": type})

    if trigger_price is not None:

        params.update({"trigger": trigger})
        params.update({"trigger_price": trigger_price})
        params.update({"reduce_only": reduce_only})

    if "market" not in type:
        params.update({"price": price})
        params.update({"post_only": post_only})
        params.update({"reject_post_only": reject_post_only})

    if otoco_config:
        params.update({"otoco_config": otoco_config})

        if linked_order_type is not None:
            params.update({"linked_order_type": linked_order_type})
        else:
            params.update({"linked_order_type": "one_triggers_other"})

        params.update({"trigger_fill_condition": "incremental"})

    return params


def send_limit_order_params(
    params: dict,
) -> str:
    """ """

    # basic params
    side = params["side"]
    instrument = params["instrument_name"]
    label_numbered = params["label"]
    size = params["size"]
    type = params["type"]
    limit_prc = params["entry_price"]

    try:
        otoco_config = params["otoco_config"]
    except:
        otoco_config = None

    try:
        linked_order_type = params["linked_order_type"]

    except:
        linked_order_type = None

    order_result = None

    if side != None:

        if type == "limit":  # limit has various state

            order_result = send_orders_params(
                side,
                instrument,
                size,
                label_numbered,
                limit_prc,
                type,
                otoco_config,
            )

        else:

            trigger_price = params["trigger_price"]
            trigger = params["trigger"]

            order_result = send_orders_params(
                side,
                instrument,
                size,
                label_numbered,
                limit_prc,
                type,
                otoco_config,
                linked_order_type,
                trigger_price,
                trigger,
            )

    return order_result


def cancel_all_orders_end_point() -> str:
    return f"private/cancel_all"

def cancel_all_orders_params() -> dict:
    return {"detailed": False}


def cancel_order_end_point() -> str:
    return f"private/cancel"


def get_cancel_order_params(
    order_id: str,
) -> dict:

    return {"order_id": order_id}


def get_api_end_point(
    endpoint: str,
    parameters: dict = None,
) -> dict:
    """
    used in data receiver deribit

    """

    private_endpoint = f"private/{endpoint}"

    params = {}
    params.update({"jsonrpc": "2.0"})
    params.update({"method": private_endpoint})
    if endpoint == "get_subaccounts":
        params.update({"params": {"with_portfolio": True}})

    if endpoint == "get_open_orders":
        end_point_params = dict(kind=parameters["kind"], type=parameters["type"])

        params.update({"params": end_point_params})

    return params


def id_numbering(
    operation: str,
    ws_channel: str,
) -> str:
    """

    id convention

    method
    subscription    3
    get             4

    auth
    public	        1
    private	        2

    instruments
    all             0
    btc             1
    eth             2

    subscription
    --------------  method      auth    seq    inst
    portfolio	        3	    1	    01
    user_order	        3	    1	    02
    my_trade	        3	    1	    03
    order_book	        3	    2	    04
    trade	            3	    1	    05
    index	            3	    1	    06
    announcement	    3	    1	    07

    get
    --------------
    currencies	        4	    2	    01
    instruments	        4	    2	    02
    positions	        4	    1	    03

    """
    id_auth = 1
    if "user" in ws_channel:
        id_auth = 9

    id_method = 0
    if "subscribe" in operation:
        id_method = 3
    if "get" in operation:
        id_method = 4
    id_item = 0
    if "book" in ws_channel:
        id_item = 1
    if "user" in ws_channel:
        id_item = 2
    if "chart" in ws_channel:
        id_item = 3
    if "index" in ws_channel:
        id_item = 4
    if "order" in ws_channel:
        id_item = 5
    if "position" in ws_channel:
        id_item = 6
    id_instrument = 0
    if "BTC" or "btc" in ws_channel:
        id_instrument = 1
    if "ETH" or "eth" in ws_channel:
        id_instrument = 2
    return int(f"{id_auth}{id_method}{id_item}{id_instrument}")
